import datetime
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import logging
import sys
import yaml
import copy
import pydantic
from pydantic.color import Color as pydColor
#from pydantic_extra_types import Color as pydColor
import typing
import pyqtgraph
import redvypr.data_packets
import redvypr.gui
import redvypr.files as files
from redvypr.widgets.pydanticConfigWidget import pydanticConfigWidget
from redvypr.redvypr_address import RedvyprAddress

_logo_file = files.logo_file
_icon_file = files.icon_file
pyqtgraph.setConfigOption('background', 'w')
pyqtgraph.setConfigOption('foreground', 'k')

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('plot_widgets')
#logger.setLevel(logging.DEBUG)
logger.setLevel(logging.INFO)

colors = ['red','blue','green','gray','yellow','purple']

class dataBufferLine(pydantic.BaseModel):
    model_config = {'extra': 'allow'}
    skip_bufferdata_when_serialized: bool = pydantic.Field(default=True, description='Do not save the buffer data when serialized')
    tdata: list = pydantic.Field(default=[])
    xdata: list = pydantic.Field(default=[])
    ydata: list = pydantic.Field(default=[])
    errordata: list = pydantic.Field(default=[])
    tdata_addr: RedvyprAddress = pydantic.Field(default=RedvyprAddress(''))
    xdata_addr: RedvyprAddress = pydantic.Field(default=RedvyprAddress(''))
    ydata_addr: RedvyprAddress = pydantic.Field(default=RedvyprAddress(''))
    errordata_addr: RedvyprAddress = pydantic.Field(default=RedvyprAddress(''))

    @pydantic.model_serializer
    def ser_model(self) -> typing.Dict[str, typing.Any]:
        if self.skip_bufferdata_when_serialized:
            return {'tdata': [],'xdata': [],'ydata': [],'errordata': [], 'skip_bufferdata_when_serialized':self.skip_bufferdata_when_serialized}
        else:
            return {'tdata': self.tdata, 'xdata': self.xdata, 'ydata': self.ydata, 'errordata': self.errordata,
                    'skip_bufferdata_when_serialized': self.skip_bufferdata_when_serialized}

class configLine(pydantic.BaseModel,extra='allow'):
    buffersize: int = pydantic.Field(default=20000,description='The size of the buffer holding the data of the line')
    numplot_max: int = pydantic.Field(default=2000, description='The number of data points to be plotted maximally')
    name: str = pydantic.Field(default='Line', description='The name of the line, this is shown in the legend, $y to use the redvypr address')
    unit: str = pydantic.Field(default='', description='The unit of the line')
    label: str = pydantic.Field(default='', description='The of the line')
    label_format: str = pydantic.Field(default='{NAME} {Y_ADDR} [{UNIT}]', description='The name of the line, this is shown in the legend, $y to use the redvypr address')
    x_addr: typing.Union[typing.Literal['$t(y)'],RedvyprAddress] = pydantic.Field(default='$t(y)', description='The realtimedata address of the x-axis, use $t(y) to automatically choose the time corresponding to the y-data')
    y_addr: RedvyprAddress = pydantic.Field(default=RedvyprAddress('/d:somedevice/k:data'), description='The realtimedata address of the x-axis')
    error_addr: RedvyprAddress = pydantic.Field(default=RedvyprAddress(''), description='The realtimedata address for an optional error band around the line')
    error_mode: typing.Literal['off', 'standard', 'factor', 'constant'] = pydantic.Field(default='off', description='')
    error_factor: float = pydantic.Field(default=1.1, description='')
    error_constant: float = pydantic.Field(default=.01, description='')
    color: pydColor = pydantic.Field(default=pydColor('red'), description='The color of the line')
    linewidth: float = pydantic.Field(default=2.0, description='The linewidth')
    linestyle: typing.Literal['SolidLine','DashLine','DotLine','DashDotLine','DashDotDotLine'] = pydantic.Field(default='SolidLine', description='The linestyle, see also https://doc.qt.io/qt-6/qt.html#PenStyle-enum')
    databuffer: dataBufferLine = pydantic.Field(default=dataBufferLine(), description='The databuffer', editable=False)
    plot_mode_x: typing.Literal['all', 'last_N_s', 'last_N_points'] = pydantic.Field(default='all', description='')
    last_N_s: float = pydantic.Field(default=60,
                                     description='Plots the last seconds, if plot_mode_x is set to last_N_s')
    last_N_points: int = pydantic.Field(default=1000,
                                        description='Plots the last points, if plot_mode_x is set to last_N_points')
    plot_every_Nth: int = pydantic.Field(default=1, description='Uses every Nth datapoint for plotting')

class configXYplot(pydantic.BaseModel):
    location: list  = pydantic.Field(default=[])
    type: str = 'XYplot'
    dt_update: float = pydantic.Field(default=0.25,description='Update time of the plot [s]')
    interactive: typing.Literal['standard', 'mouse'] = pydantic.Field(default='standard',description='Interactive modes')
    backgroundcolor: pydColor = pydantic.Field(default=pydColor('lightgray'),description='Backgroundcolor')
    bordercolor: pydColor = pydantic.Field(default=pydColor('lightgray'), description='Bordercolor')
    show_legend: bool = pydantic.Field(default=True, description='Show legend (True) or hide (False)')
    show_units: bool = pydantic.Field(default=True, description='Add the unit of the y-data to the legend, queried from the datakey')
    datetick_x: bool = pydantic.Field(default=True, description='x-axis is a date axis')
    datetick_y: bool = pydantic.Field(default=True, description='y-axis is a date axis')
    title: str = pydantic.Field(default='', description='')
    name: str = pydantic.Field(default='', description='The name of the plotWidget')
    xlabel: str = pydantic.Field(default='', description='')
    ylabel: str = pydantic.Field(default='', description='')
    lines: typing.Optional[typing.List[configLine]] = pydantic.Field(default=[configLine()], editable=True)
    plot_mode_x: typing.Literal['all', 'last_N_s', 'last_N_unit'] = pydantic.Field(default='all', description='')
    last_N_s: float = pydantic.Field(default=10,
                                     description='Plots the last seconds, if plot_mode_x is set to last_N_s')
    last_N_points: int = pydantic.Field(default=1000,
                                        description='Plots the last points, if plot_mode_x is set to last_N_points')
    automatic_subscription: bool = pydantic.Field(default=True,
                                                  description='subscribes automatically the adresses of the lines at the host device')

# config_template_graph['description'] = description_graph
class XYPlotWidget(QtWidgets.QFrame):
    """ Widget is plotting realtimedata using the pyqtgraph functionality

    """
    closing = QtCore.pyqtSignal()  # Signal notifying that a subscription changed

    def __init__(self, config=None, redvypr_device=None, add_line=True, loglevel=logging.INFO):
        """

        """
        funcname = __name__ + '.init():'
        super(QtWidgets.QFrame, self).__init__()
        self.device = redvypr_device
        if self.device is not None:
            self.redvypr = self.device.redvypr
        else:
            self.redvypr = None
        self.logger = logging.getLogger('XYplot')
        self.logger.setLevel(loglevel)
        self.description = 'XY plot'

        self.x_min = 0
        self.x_max = 0
        if (config == None):  # Create a config from the template
            self.config = configXYplot()
        else:
            self.config = config

        if add_line == False:
            try:
                self.config.lines.pop(0)
            except:
                logger.debug(funcname, exc_info=True)

        # self.logger.debug('plot widget config {:s}'.format(str(self.config)))
        backcolor = str(self.config.backgroundcolor.as_rgb())
        bordercolor = str(self.config.bordercolor.as_rgb())
        # logger.debug(funcname + 'backcolor {:s}'.format(str(backcolor)))
        style = "background-color : {:s};border : 1px solid {:s};".format(backcolor, bordercolor)
        logger.debug(funcname + 'Style: {:s}'.format(str(style)))
        self.setStyleSheet(style)
        self.layout = QtWidgets.QVBoxLayout(self)
        self.create_widgets()
        self.apply_config()
        self.xAxisLimitsChanged()

    def create_widgets(self):

        """
        Creates the configuration

        Returns:

        """
        funcname = __name__ + '.create_widgets()'
        config = self.config
        self.logger.debug(funcname)
        i = 0
        if True:
            self.logger.debug(funcname + ': Adding plot ')  # + str(config))
            try:
                title = config.title
            except:
                title = "Plot {:d}".format(i)

            try:
                name = config.name
            except:
                name = "Plot {:d}".format(i)

            # https://stackoverflow.com/questions/44402399/how-to-disable-the-default-context-menu-of-pyqtgraph
            plot = pyqtgraph.PlotWidget(title=title, name=name)
            plot.plotItem.vb.menu.clear()
            # General config
            configAction = plot.plotItem.vb.menu.addAction('General config')
            configAction.triggered.connect(self.pyqtgraphConfigAction)
            # X-Axis
            xMenu = plot.plotItem.vb.menu.addMenu('X-Axis')
            xAction = QtWidgets.QWidgetAction(self)
            # Create X-Menu
            xMenuWidget = QtWidgets.QWidget()
            xMenuWidget_layout = QtWidgets.QVBoxLayout(xMenuWidget)
            self._xaxis_radio_auto = QtWidgets.QRadioButton('Autoscale')
            self._xaxis_radio_lasts = QtWidgets.QRadioButton('Last N-Seconds')

            if self.config.plot_mode_x == 'all':
                self._xaxis_radio_auto.setChecked(True)
            else:
                self._xaxis_radio_lasts.setChecked(True)

            self._xaxis_radio_auto.toggled.connect(self.xAxisLimitsChanged)
            self._xaxis_spin_lasts = QtWidgets.QDoubleSpinBox()
            self._xaxis_spin_lasts.setValue(self.config.last_N_s)
            self._xaxis_spin_lasts.setMinimum(0)
            self._xaxis_spin_lasts.setMaximum(1e12)
            self._xaxis_spin_lasts.valueChanged.connect(self.xAxisLimitsChanged)
            xMenuWidget_layout.addWidget(self._xaxis_radio_auto)
            xMenuWidget_layout.addWidget(self._xaxis_radio_lasts)
            xMenuWidget_layout.addWidget(self._xaxis_spin_lasts)
            xAction.setDefaultWidget(xMenuWidget)
            xMenu.addAction(xAction)
            #xMenu.triggered.connect(self.pyqtgraphXMenuAction)
            # Y-Axis
            yMenu = plot.plotItem.vb.menu.addMenu('Y-Axis')
            #yMenu.triggered.connect(self.pyqtgraphYMenuAction)
            # Line config menu
            menu = plot.plotItem.vb.menu.addMenu('Line config')
            self.lineMenu = menu
            # Clear buffer
            bufferMenu = plot.plotItem.vb.menu.addMenu('Clear buffer')
            self.bufferMenu = bufferMenu
            # Remove menu
            removeMenu = plot.plotItem.vb.menu.addMenu('Remove line')
            self.removeMenu = removeMenu
            # Add line
            addLineAction = plot.plotItem.vb.menu.addAction('Add line')
            addLineAction.triggered.connect(self.pyqtgraphAddLineAction)

            # plot = redvyprPlotWidget(title=title,name=name)
            plot.register(name=name)
            # Add a legend
            legend = plot.addLegend()
            if config.interactive.lower() == 'mouse':
                self.logger.debug('Adding interactive mouse (vertical line)')
                plot.scene().sigMouseMoved.connect(self.mouseMoved)
                plot.scene().sigMouseClicked.connect(self.mouseClicked)
                self.vb = plot.plotItem.vb
                # Add a vertical line
                color = QtGui.QColor(200, 100, 100)
                linewidth = 2.0
                pen = pyqtgraph.mkPen(color, width=linewidth)
                self.vLineMouse = pyqtgraph.InfiniteLine(angle=90, movable=False, pen=pen)
                # print('Set pen 3')
                plot.addItem(self.vLineMouse, ignoreBounds=True)

            self.layout.addWidget(plot)
            self.plotWidget = plot
            self.legendWidget = legend
            # plot_dict = {'widget': plot, 'lines': []}

    def xAxisLimitsChanged(self):
        funcname = __name__ + '.xAxisLimitsChanged():'
        logger.debug(funcname)
        self.config.last_N_s = self._xaxis_spin_lasts.value()
        if self._xaxis_radio_auto.isChecked():
            logger.debug(funcname + 'Enabling auto scaling')
            self.config.plot_mode_x = 'all'
        if self._xaxis_radio_lasts.isChecked():
            logger.debug(funcname + 'Enabling last s range')
            self.config.plot_mode_x = 'last_N_s'

        # Update if it is existing already ...
        try:
            self.plotWidget
            self.applyPlotRanges()
        except:
            pass

    def applyPlotRanges(self):
        funcname = __name__ + '.applyPlotRanges():'
        logger.debug(funcname)
        if self.config.plot_mode_x == 'all':
            self.plotWidget.enableAutoRange(axis='x')
            self.plotWidget.setAutoVisible(x=True)
        elif self.config.plot_mode_x == 'last_N_s':
            xmin = self.x_max - self.config.last_N_s
            xmax = self.x_max
            self.plotWidget.setXRange(xmin, xmax)

        self.plotWidget.enableAutoRange(axis='y')
        self.plotWidget.setAutoVisible(y=True)

    def pyqtgraphXMenuAction(self):
        funcname = __name__ + '.pyqtgraphXMenuAction()'
        logger.debug(funcname)

    def pyqtgraphYMenuAction(self):
        funcname = __name__ + '.pyqtgraphYMenuAction()'
        logger.debug(funcname)

    def pyqtgraphConfigAction(self):
        funcname = __name__ + '.pyqtgraphConfigAction()'
        logger.debug(funcname)
        self.ConfigWidget = pydanticConfigWidget(self.config, configname='new line',exclude=['lines'], redvypr=self.redvypr)
        self.ConfigWidget.config_changed_flag.connect(self.apply_config)
        self.ConfigWidget.show()

    def pyqtgraphRemLineAction(self):
        funcname = __name__ + '.pyqtgraphRemLineAction()'
        logger.debug(funcname)
        index_remove = self.sender()._iline
        self.config.lines.pop(index_remove)
        self.sender()._line._lineplot.clear()
        self.apply_config()

    def pyqtgraphAddLineAction(self):
        funcname = __name__ + '.pyqtgraphAddLineAction()'
        logger.debug(funcname)
        newline = configLine()
        self.__newline = newline
        self.addLineConfigWidget = pydanticConfigWidget(newline, configname='new line', redvypr=self.device.redvypr)
        self.addLineConfigWidget.setWindowTitle('Add line')
        self.addLineConfigWidget.config_editing_done.connect(self.pyqtgraphAddLineDone)
        self.addLineConfigWidget.show()

    def pyqtgraphAddLineDone(self):
        if self.__newline is not None:
            self.config.lines.append(self.__newline)
        self.apply_config()

    def pyqtgraphLineAction(self):
        """ Function is called whenever a line is configured

        """
        funcname = __name__ + '.pyqtgraphLineAction()'
        self.logger.debug(funcname)
        config_mode = self.sender().text()
        if 'address' in config_mode.lower():
            #print('Address')
            linename = self.sender()._linename
            lineConfig = self.sender()._line
            self.lineConfigWidget = redvypr.widgets.redvyprAddressWidget.datastreamWidget(redvypr=self.redvypr, device=self.device)
            self.lineConfigWidget.setWindowTitle('Y-Address for {}'.format(linename))
            self.lineConfigWidget.apply.connect(self.apply_config_address)
            self.lineConfigWidget._line = self.sender()._line
            self.lineConfigWidget.show()

        elif 'color' in config_mode.lower():
            #print('Color')
            linename = self.sender()._linename
            lineConfig = self.sender()._line
            self.lineConfigWidget = QtWidgets.QColorDialog()
            self.lineConfigWidget._line = lineConfig
            #self.lineConfigWidget.currentColorChanged.connect(self.apply_config_color)
            self.lineConfigWidget.accepted.connect(self.apply_config_color)
            self.lineConfigWidget.show()

        elif 'settings' in config_mode.lower():
            linename = self.sender()._linename
            lineConfig = self.sender()._line
            self.lineConfigWidget = pydanticConfigWidget(lineConfig, configname=linename, redvypr=self.device.redvypr)
            self.lineConfigWidget.setWindowTitle('Config {}'.format(linename))
            self.lineConfigWidget.config_editing_done.connect(self.apply_config)
            self.lineConfigWidget.show()

    def pyqtgraphBufferAction(self):
        """ Function is called whenever a clear buffer action is called

        """
        funcname = __name__ + '.pyqtgraphBufferAction()'
        self.logger.debug(funcname)
        self.clear_buffer()

    def mouseMoved(self, evt):
        """Function if mouse has been moved
        """
        pos = (evt.x(), evt.y())
        mousePoint = self.vb.mapSceneToView(evt)
        # print('mouse moved',mousePoint.x())
        self.vLineMouse.setPos(mousePoint.x())
        # for vline in self.vlines:
        #    vline.setPos(mousePoint.x())

    def mouseClicked(self, evt):
        # col =
        # col = pg.mkPen(0.5,width=3)
        # colsymbol = pg.mkPen(color=QtGui.QColor(150,150,150),width=4)
        # print('Clicked: ' + str(evt.scenePos()))
        pass

    def set_title(self, title):
        """
        Sets the title of the plot
        """
        funcname = __name__ + '.set_title()'
        self.logger.debug(funcname)
        self.config.title = title
        self.plotWidget.setTitle(title)

    def set_line(self, index, y_addr, name='$y', x_addr='$t(y)', color=[255, 0, 0], linewidth=1, bufsize=2000):
        """
        Sets the line at index to the parameters
        """
        self.add_line(y_addr=y_addr, name=name, x_addr=x_addr, color=color, linewidth=linewidth, bufsize=bufsize, index=index)

    def add_line(self, y_addr, x_addr='$t(y)', name='$y', error_addr='', color=None, linewidth=1, bufsize=20000, numplot=2000, index=None):
        """
        Adds a line to the plot
        """
        funcname = __name__ + '.add_line()'
        self.logger.debug(funcname)
        if isinstance(y_addr, redvypr.RedvyprAddress):
            self.logger.debug('Redvypr y-address')
            y_addr = y_addr.address_str
        if isinstance(x_addr, redvypr.RedvyprAddress):
            self.logger.debug('Redvypr x-address')
            x_addr = x_addr.address_str
        if isinstance(error_addr, redvypr.RedvyprAddress):
            self.logger.debug('Redvypr error-address')
            error_addr = error_addr.address_str
        #print('add line',y_addr,color)
        if color is None: # No color defined, take color from the colors list
            nlines = len(self.config.lines) - 1
            colind = nlines % len(colors)
            color = colors[colind]
            self.logger.debug('Color')

        if index is None:
            rconfig = configLine()
            self.config.lines.append(rconfig)
            index = len(self.config.lines) - 1

        self.config.lines[index].buffersize = bufsize
        self.config.lines[index].numplot_max = numplot
        self.config.lines[index].linewidth = linewidth
        #print('Using color',color)
        color_tmp = pydColor(color)
        self.config.lines[index].color = color_tmp
        self.config.lines[index].x_addr = x_addr
        self.config.lines[index].y_addr = y_addr
        self.config.lines[index].error_addr = error_addr
        self.config.lines[index].name = name
        # Add the address as well to the data
        self.config.lines[index].databuffer.xdata_addr = x_addr
        self.apply_config()

    def construct_labelname(self, line, labelformat=None):
        name = line.name
        unit = line.unit
        x_addr = RedvyprAddress(line.x_addr).address_str_explicit
        y_addr = RedvyprAddress(line.y_addr).address_str_explicit
        if labelformat == None:
            labelname = line.label_format
            labelname = labelname.format(NAME=name,UNIT=unit,X_ADDR=x_addr,Y_ADDR=y_addr)
        self.logger.debug('Labelname {}'.format(labelname))
        return labelname

    def apply_config_address(self,address_dict):
        funcname = __name__ + '.apply_config_address():'
        self.logger.debug(funcname)
        line = self.sender()._line
        #print('Line',line,line.y_addr)
        line.y_addr = address_dict['datastream_str']
        #print('Line config',line.confg)
        self.apply_config()

    def apply_config_color(self):
        funcname = __name__ + 'apply_config_color():'
        logger.debug(funcname)
        try:
            color = self.sender().currentColor()  # ComboBox
            color1 = color.getRgb()
            color_tmp = (color1[0], color1[1], color1[2])
            rint = int(color1[0] * 255)
            gint = int(color1[1] * 255)
            bint = int(color1[2] * 255)
            data = pydColor(color_tmp)
            self.sender()._line.color = data
            self.apply_config()
        except:
            logger.info('Could not set color',exc_info=True)

    def apply_config_linewidth(self):
        spinbox = self.sender()
        spinbox._line.linewidth = spinbox.value()
        self.apply_config()

    def apply_config_linestyle(self):
        combobox = self.sender()
        combobox._line.linestyle = combobox.currentText()
        self.apply_config()

    def apply_config(self):
        """
        Function is called by the initialization or after the configuration was changed

        Returns:

        """
        funcname = __name__ + '.apply_config():'
        self.logger.debug(funcname)
        # Recreate the lineMenu
        self.lineMenu.clear()
        self.removeMenu.clear()
        self.bufferMenu.clear()
        # Add a clear buffer action
        bufferAction = self.bufferMenu.addAction('All lines')
        bufferAction.triggered.connect(self.pyqtgraphBufferAction)

        plot = self.plotWidget
        # Title
        title = self.config.title
        plot.setTitle(title)

        try:
            datetick_x = self.config.datetick_x
        except:
            datetick_x = False

        if datetick_x:
            self.logger.debug(funcname + 'Datetick')
            axis = pyqtgraph.DateAxisItem(orientation='bottom', utcOffset=0)
            plot.setAxisItems({"bottom": axis})
        else:
            self.logger.debug(funcname + ' No Datetick')
            axis = pyqtgraph.AxisItem(orientation='bottom')
            plot.setAxisItems({"bottom": axis})

        # Label
        # If a xlabel is defined
        try:
            plot.setLabel('left', self.config.ylabel)
        except:
            pass

        # If a ylabel is defined
        try:
            plot.setLabel('bottom', self.config.xlabel)
        except:
            pass

        # Legend update
        self.legendWidget.clear()
        # Add lines with the actual data to the graph
        for iline, line in enumerate(self.config.lines):
            self.logger.debug(funcname + ': Updating line {:d}'.format(iline))
            # Creating the correct addresses first
            try:
                self.logger.debug('Line {},{}'.format(line,iline))
                x_addr = line.x_addr
                y_addr = line.y_addr
                if len(y_addr) >= 0:
                    y_raddr = redvypr.RedvyprAddress(y_addr)
                    #print('x_addr', x_addr)
                    #print('y_addr', y_addr)
                    if (x_addr == '$t(y)'):
                        self.logger.debug(funcname + ' Using time variable of y')
                        #xtmp = redvypr.data_packets.modify_addrstr(y_raddr.address_str, datakey='t')
                        ## print('xtmp',xtmp)
                        #x_raddr = redvypr.redvypr_address(xtmp)
                        x_raddr = redvypr.RedvyprAddress(y_addr, datakey='t')
                    else:
                        x_raddr = redvypr.RedvyprAddress(x_addr)

                    #print('x_addrnew', x_addr,x_raddr)
                    # These attributes are used in plot.Device.connect_devices to actually subscribe to the fitting devices
                    line._x_raddr = x_raddr
                    line._y_raddr = y_raddr

                # Error address
                error_addr = None
                if line.error_mode != 'off':
                    self.logger.debug('Error mode 1')
                    error_raddr = redvypr.RedvyprAddress(line.error_addr)
                    line._error_raddr = error_raddr
                    errorplot = pyqtgraph.ErrorBarItem(name=line._name_applied,pen=pen)
                    line._errorplot = errorplot  # Add the line as an attribute to the configuration
                    errorplot._line_config = line
                    plot.addItem(errorplot)
                else:
                    self.logger.debug('Error mode 2')
                    line._error_raddr = redvypr.RedvyprAddress('')
                    line._errorplot = None
                    # print('Set pen 1')
            except:
                self.logger.debug(funcname + 'Could not update addresses',exc_info=True)
                continue

            # print('Line',line)
            # FLAG_HAVE_LINE = False
            self.get_metadata_for_line(line) # This is updating the unit
            labelname = self.construct_labelname(line)
            line.label = labelname
            linename = "Line {}: {}".format(iline, labelname)
            linename_menu = "Line {}".format(iline)

            # check if we have already a lineplot, if yes, dont bother
            try:
                self.logger.debug(funcname + 'We have already a lineplot, no update')
                lineplot = line._lineplot
            except:
                self.logger.debug(funcname + 'Creating new line with labelname {}'.format(labelname), exc_info=True)
                lineplot = pyqtgraph.PlotDataItem(name=labelname)
                line._lineplot = lineplot  # Add the line as an attribute to the configuration
                plot.addItem(lineplot)

            try:
                line._name_applied = line.name
            except:
                line._name_applied = 'line {:d}'.format(iline)

            lineplot._line_config = line
            line._tlastupdate = 0
            try:
                #print('Color',self.config.lines[iline].color.as_rgb())
                color = redvypr.gui.get_QColor(self.config.lines[iline].color)
                # print('Set pen 2')
                linewidth = self.config.lines[iline].linewidth
                linestyle = self.config.lines[iline].linestyle
                style=eval('QtCore.Qt.' + linestyle)
                # print('Set pen 3')
                #color = QtGui.QColor(200, 100, 100)
                #print('COLOR!!!!', color, type(color), linewidth)
                pen = pyqtgraph.mkPen(color, width=float(linewidth),style=style)
                line._lineplot.setPen(pen)
                # Check if the addresses changed and clear the buffer if necessary
                if line.databuffer.xdata_addr == line.x_addr:
                    self.logger.debug('Address did not change')
                else:
                    pass
                    # TODO, here an unsubscribe and resubscribe would be better
                    #self.logger.debug('Address changed, clearing buffer')
                    #self.clear_buffer(line)

                self.logger.debug(funcname + ' Setting the name')
                try:
                    self.legendWidget.removeItem(line._lineplot)
                except:
                    self.logger.debug(funcname + ' Could not remove line')
                self.legendWidget.addItem(line._lineplot, line.label)
                self.logger.debug('Setting the data')
                line._lineplot.setData(name=line.label,x=[],y=[])
                if self.config.automatic_subscription:
                    self.logger.debug(funcname + 'Subscribing to x address {}'.format(x_raddr))
                    self.device.subscribe_address(x_raddr)
                    self.logger.debug(funcname + 'Subscribing to y address {}'.format(y_raddr))
                    self.device.subscribe_address(y_raddr)
                    if line._errorplot is not None:
                        self.logger.debug(funcname + 'Subscribing to error address {}'.format(error_raddr))
                        self.device.subscribe_address(error_raddr)

                # Create the menu to config the line
                #lineMenuline = QtWidgets.QMenu(linename_menu,self)
                lineMenuline = QtWidgets.QMenu(linename_menu, self.lineMenu)
                lineAction = lineMenuline.addAction('Y-Address')
                lineAction._line = line
                lineAction._iline = iline
                lineAction._linename = linename
                lineAction.triggered.connect(self.pyqtgraphLineAction)

                # Create a widget for the linewidth
                linewidthAction = QtWidgets.QWidgetAction(lineMenuline)
                # Create add XYPlot-Menu
                linewidthWidget = QtWidgets.QWidget()
                linewidthWidget_layout = QtWidgets.QVBoxLayout(linewidthWidget)
                linewidthSpinBox = QtWidgets.QDoubleSpinBox()
                linewidthSpinBox.setValue(line.linewidth)
                linewidthSpinBox._line = line
                linewidthSpinBox.valueChanged.connect(self.apply_config_linewidth)
                # Add a widget for the linestyle
                linewidthWidget_layout.addWidget(QtWidgets.QLabel('Linewidth'))
                linewidthWidget_layout.addWidget(linewidthSpinBox)
                linewidthWidget_layout.addWidget(QtWidgets.QLabel('Linestyle'))
                linestyleComboBox = QtWidgets.QComboBox()
                linestyleComboBox._line = line
                linewidthWidget_layout.addWidget(linestyleComboBox)

                styles_tmp = typing.get_type_hints(line, include_extras=True)['linestyle']
                styles_tmp = typing.get_args(styles_tmp)
                i_style_set = 0
                for i_s,s in enumerate(styles_tmp):
                    linestyleComboBox.addItem(s)
                    if s == line.linestyle:
                        i_style_set = i_s

                linestyleComboBox.setCurrentIndex(i_style_set)
                linestyleComboBox.currentTextChanged.connect(self.apply_config_linestyle)
                #print('Styles',styles_tmp)

                linewidthAction.setDefaultWidget(linewidthWidget)
                lineMenuline.addAction(linewidthAction)

                colorAction = lineMenuline.addAction('Linecolor')
                colorAction._line = line
                colorAction._iline = iline
                colorAction._linename = linename
                colorAction.triggered.connect(self.pyqtgraphLineAction)

                lineAction = lineMenuline.addAction('All settings')
                lineAction._line = line
                lineAction._iline = iline
                lineAction._linename = linename
                lineAction.triggered.connect(self.pyqtgraphLineAction)
                self.lineMenu.addMenu(lineMenuline)
                # remove line menu
                # Remove line
                lineAction = self.removeMenu.addAction(linename_menu)
                lineAction._line = line
                lineAction._iline = iline
                lineAction._linename = linename
                lineAction.triggered.connect(self.pyqtgraphRemLineAction)
            except Exception as e:
                self.logger.debug('Exception config lines: {:s}'.format(str(e)),exc_info=True)
                raise ValueError('')

        self.logger.debug(funcname + ' done.')


    def get_metadata_for_line(self, line):
        funcname = __name__ + '.get_metadata_for_line():'
        iline = self.config.lines.index(line)
        self.logger.debug(funcname + 'Getting metadata for {}'.format(line._y_raddr))
        try:
            line._metadata  # metadata found, doing nothing
        except:
            line._metadata = self.device.get_metadata(line._y_raddr)
            self.logger.debug(funcname + ' Datakeyinfo {:s}'.format(str(line._metadata)))
            try:
                unit = line._metadata['unit']
                self.logger.debug(funcname + 'Found a unit for {}'.format(line._y_raddr))
            except:
                self.logger.debug(funcname + 'Did not find a unit', exc_info=True)
                unit = None

            if unit is not None:
                self.config.lines[iline].unit = unit
                return True

        return None

    def clear_buffer(self, line=None):
        """ Clears the buffer of all lines
        """
        if line is not None:
            lines = [line]
        else:
            lines = self.config.lines

        for iline, line_tmp in enumerate(lines):
            line_tmp.databuffer = dataBufferLine()
            # Set the data
            line_tmp._lineplot.setData(x=line_tmp.databuffer.xdata, y=line_tmp.databuffer.ydata)

    def get_data(self, xlim):
        """
        Gets the data of the buffer in the limits of xlim
        """
        funcname = __name__ + '.get_data():'
        data = []
        tdata = []
        xdata = []
        ydata = []
        for iline, line in enumerate(self.config.lines):
            #print('Line', line)
            #print('Type Line', type(line))
            #line_dict = line.line_dict
            #line = line_dict['line']  # The line to plot
            #config = line_dict['config']  # The line to plot
            t = np.asarray(line.databuffer.tdata)  # The line to plot
            x = np.asarray(line.databuffer.xdata)  # The line to plot
            y = np.asarray(line.databuffer.ydata)  # The line to plot
            err = np.asarray(line.databuffer.errordata)  # The line to plot
            ind = (x > xlim[0]) & (x < xlim[1])
            tdata_tmp = t[ind]
            xdata_tmp = x[ind]
            ydata_tmp = y[ind]
            data.append({'x': xdata_tmp, 'y': ydata_tmp, 't': tdata_tmp})
            #tdata.append(tdata_tmp)
            #xdata.append(xdata_tmp)
            #ydata.append(ydata_tmp)
            self.logger.debug(funcname + ' Got data of length {:d}'.format(len(tdata_tmp)))
            # print('get_data',datetime.datetime.utcfromtimestamp(tdata_tmp[0]),datetime.datetime.utcfromtimestamp(xdata_tmp[0]))

        return data

    def __get_data_for_line(self, line):
        """
        Provides the data for the line to plot
        :return:

        plot_mode_x: typing.Literal['all', 'last_N_s', 'last_N_points'] = pydantic.Field(default='all', description='')
    last_N_s: float = pydantic.Field(default=60,
                                     description='Plots the last seconds, if plot_mode_x is set to last_N_s')
    last_N_points: int = pydantic.Field(default=200,
                                        description='Plots the last points, if plot_mode_x is set to last_N_points')
    plot_every_Nth: int = pydantic.Field(default=1, description='Uses every Nth datapoint for plotting')

        """

        if line.plot_mode_x == 'all':
            ttmp = line.databuffer.tdata[::line.plot_every_Nth]
            xtmp = line.databuffer.xdata[::line.plot_every_Nth]
            ytmp = line.databuffer.ydata[::line.plot_every_Nth]
            errtmp = line.databuffer.errordata[::line.plot_every_Nth]
        elif line.plot_mode_x == 'last_N_s':
            ttmp = line.databuffer.tdata[::line.plot_every_Nth]
            xtmp = line.databuffer.xdata[::line.plot_every_Nth]
            ytmp = line.databuffer.ydata[::line.plot_every_Nth]
            errtmp = line.databuffer.errordata[::line.plot_every_Nth]
            # Find the index
            tnow = ttmp[-1]
            tsearch = tnow - line.last_N_s
            ttmp_ar = np.asarray(ttmp)
            iplot = np.where(ttmp_ar >= tsearch)[0]
            if len(iplot)>0:
                istart = iplot[0]
                ttmp = ttmp[istart:]
                xtmp = xtmp[istart:]
                ytmp = ytmp[istart:]
                errtmp = errtmp[istart:]
            else:
                ttmp = []
                xtmp = []
                ytmp = []
                errtmp = []

        elif line.plot_mode_x == 'last_N_points':
            ttmp = line.databuffer.tdata[::line.plot_every_Nth][-line.last_N_points:]
            xtmp = line.databuffer.xdata[::line.plot_every_Nth][-line.last_N_points:]
            ytmp = line.databuffer.ydata[::line.plot_every_Nth][-line.last_N_points:]
            errtmp = line.databuffer.errordata[::line.plot_every_Nth][-line.last_N_points:]
        else:
            pass

        # Reduce the number of points, if they are more than numplot_max
        x = xtmp[-line.numplot_max:]
        y = ytmp[-line.numplot_max:]
        err = errtmp[-line.numplot_max:]
        # pw.setXRange(min(x[:ind]),max(x[:ind]))
        return [x,y,err]

    def closeEvent(self, event):
        #print('Close event')
        self.closing.emit()

    def update_plot(self, data):
        """ Updates the plot based on the given data
        """
        funcname = self.__class__.__name__ + '.update_plot():'
        #self.logger.debug(funcname + 'Update {}'.format(len(self.config.lines)))
        tnow = time.time()
        # Create a redvypr datapacket
        rdata = redvypr.data_packets.Datapacket(data)
        # print(funcname + 'got data',data,tnow)
        try:
            # Check if the device is to be plotted
            # Loop over all lines
            for iline, line in enumerate(self.config.lines):
                line.__newdata = False
                error_raddr = line._error_raddr
                if len(self.config.lines)>1:
                    #print('device',data['_redvypr']['device'])
                    #print('data',data)
                    #print('line',line)
                    #print('line A', line._x_raddr, (data in line._x_raddr))
                    #print('line B', line._y_raddr, (data in line._y_raddr))
                    #print('fdsfsfsd',(data in line._x_raddr) and (data in line._y_raddr))
                    pass

                #print('data',data)
                #print('line._x_raddr',line._x_raddr,data in line._x_raddr)
                #print('line._y_raddr', line._y_raddr, data in line._y_raddr)
                if (data in line._x_raddr) and (data in line._y_raddr):
                    pw = self.plotWidget  # The plot widget
                    #if len(self.config.lines) > 1:
                    #    print('Databuffer',line.databuffer)
                    tdata = line.databuffer.tdata  # The line to plot
                    xdata = line.databuffer.xdata  # The line to plot
                    ydata = line.databuffer.ydata  # The line to plot
                    # data can be a single float or a list, if its a list add it item by item
                    newt = rdata['_redvypr']['t']  # Add also the time of the packet
                    newx = rdata[line._x_raddr.datakey]
                    newy = rdata[line._y_raddr.datakey]

                    #if len(self.config.lines) > 0:
                    #    print('data xy plotwidget',rdata)
                    #    print('newx datakey', line._x_raddr.datakey)
                    #    print('newx', newx)

                    if (type(newx) is not list):
                        newx = [newx]
                    if (type(newy) is not list):
                        newy = [newy]

                    if  line.error_mode != 'off':
                        error_mode: typing.Literal['off','standard', 'factor', 'constant'] = pydantic.Field(
                            default='standard', description='')
                        error_factor: float = pydantic.Field(default=1.1, description='')
                        error_constant: float = pydantic.Field(default=.01, description='')
                        #print('errordata',error_raddr.datakey)
                        if len(line.error_addr) > 0 and line.error_mode == 'standard':
                            #logger.debug('Error standard')
                            newerror = data[error_raddr.datakey]
                            if (type(newerror) is not list):
                                newerror = [newerror]
                            #print('newerror',newerror)
                        elif line.error_mode == 'factor':
                            #print('Error factor')
                            errdata = np.asarray(newy)
                            errdata_factor = errdata * line.error_factor - errdata.mean()
                            newerror = errdata_factor.tolist()
                        elif line.error_mode == 'constant':
                            #print('Error constant')
                            newerror = [line.error_constant] * len(newx)
                    else:
                        newerror = [0]

                    if (len(newx) != len(newy)):
                        self.logger.warning(
                            'lengths of x and y data different (x:{:d}, y:{:d})'.format(len(newx), len(newy)))
                        return

                    for inew in range(len(newx)):  # TODO this can be optimized using indices instead of a loop
                        line.databuffer.tdata.append( float(newt) )
                        line.databuffer.xdata.append( float(newx[inew]) )
                        line.databuffer.ydata.append( float(newy[inew]) )
                        line.databuffer.errordata.append(float(newerror[inew]) )
                        while len(line.databuffer.tdata) > line.buffersize:
                            line.databuffer.tdata.pop(0)
                            line.databuffer.xdata.pop(0)
                            line.databuffer.ydata.pop(0)
                            line.databuffer.errordata.pop(0)

                        line.__newdata = True

                        #tdata = np.roll(tdata, -1)
                        #xdata = np.roll(xdata, -1)
                        #ydata = np.roll(ydata, -1)
                        #tdata[-1] = float(newt)
                        #xdata[-1] = float(newx[inew])
                        #ydata[-1] = float(newy[inew])
                        #line.databuffer.tdata = tdata
                        #line.databuffer.xdata = xdata
                        #line.databuffer.ydata = ydata

                    # Show the unit in the legend, if wished by the user, and we have access to the device that can give us the metainformation
                    if (self.config.show_units) and (self.device is not None):
                        #self.logger.debug(funcname + 'Getting metadata for {}'.format(line._y_raddr))
                        try:
                            line._metadata  # metadata found, doing nothing
                        except:
                            line._metadata = self.device.get_metadata(line._y_raddr)
                            self.logger.debug(funcname + ' Datakeyinfo {:s}'.format(str(line._datakeys)))
                            try:
                                unit = line._metadata['unit']
                                self.logger.debug(funcname + 'Found a unit for {}'.format(line._y_raddr))
                            except:
                                self.logger.debug(funcname + 'Did not find a unit',exc_info=True)
                                unit = None

                            if unit is not None:
                                self.config.lines[iline].unit = unit
                                self.apply_config()

            # Update the lines plot
            something_updated = False
            for iline, line in enumerate(self.config.lines):
                tlastupdate = line._tlastupdate  # The time the plot was last updated
                # Check if an update of the plot shall be done, or if only the buffer is updated
                dt = tnow - tlastupdate
                if dt > self.config.dt_update:
                    update = True
                else:
                    update = False
                    # print('no update')

                #if len(self.config.lines) > 1:
                #    print('Update',update,line.__newdata)
                if update and line.__newdata:  # We could check here if data was changed above the for given line
                    line._tlastupdate = tnow
                    try:
                        [x,y,err]= self.__get_data_for_line(line)
                        #print('x',x,'err',err)
                        line._lineplot.setData(x=x, y=y)
                        self.x_min = min(self.x_min,min(x))
                        self.x_max = max(self.x_max, max(x))
                        something_updated = True
                        if line._errorplot is not None:
                            beamwidth = None
                            line._errorplot.setData(x=np.asarray(x), y=np.asarray(y), top=np.asarray(err) * 1,
                                                    bottom=np.asarray(err) * 1, beam=beamwidth)

                    except:
                        self.logger.info('Could not update line',exc_info=True)

            if something_updated:
                try:
                    # Check if ranges need to be changed
                    if self.config.plot_mode_x == 'last_N_s':
                        xmin = self.x_max - self.config.last_N_s
                        xmax = self.x_max
                        self.plotWidget.setXRange(xmin,xmax)
                except:
                    logger.info(funcname,exc_info=True)

            if len(self.config.lines) > 1:
                pass
                #print('DONE DONE DONE')
        except:
            self.logger.debug('Could not update data', exc_info=True)


import datetime
import redvypr.data_packets as data_packets
import numpy
import logging
import sys
import zoneinfo
from PyQt5 import QtWidgets, QtCore, QtGui
import pydantic
import typing
from redvypr.devices.plot import XYplotWidget
from redvypr.devices.plot import plot_widgets
from redvypr.redvypr_address import RedvyprAddress
from redvypr.devices.sensors.calibration.calibration_models import calibration_HF, calibration_NTC
import redvypr.gui as gui
from .sensorWidgets import sensorCoeffWidget, sensorConfigWidget


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('heatflow_sensors')
logger.setLevel(logging.DEBUG)


class parameter_DHFS50(pydantic.BaseModel):
    HF: calibration_HF = calibration_HF(parameter='HF')
    NTC0: calibration_NTC = calibration_NTC(parameter='NTC0')
    NTC1: calibration_NTC = calibration_NTC(parameter='NTC1')
    NTC2: calibration_NTC = calibration_NTC(parameter='NTC2')

class sensor_DHFS50(pydantic.BaseModel):
    description: str = 'Digital heat flow sensor DHFS50'
    sensor_type: typing.Literal['DHFS50'] = 'DHFS50'
    sensor_model: str = 'DHFS50'
    sensor_configuration: str = 'DHFS50-1m'
    parameter: parameter_DHFS50 = parameter_DHFS50()
    sn: str = pydantic.Field(default='', description='The serial number and/or MAC address of the sensor')
    comment: str = pydantic.Field(default='')
    convert_rawdata: bool = pydantic.Field(default=True, description='Uses the calibrations of the parameter to convert rawdata into meaningful units')
    avg_data: list = pydantic.Field(default=[60, 300], description='List of averaging intervals')
    show_rawdata: bool = pydantic.Field(default=True, description='Show the raw data in the gui')
    show_convdata: bool = pydantic.Field(default=True, description='Show the converted data in the gui')

class channels_HFV4CH(pydantic.BaseModel):
    C0: calibration_HF = calibration_HF(parameter='C0')
    C1: calibration_HF = calibration_HF(parameter='C1')
    C2: calibration_HF = calibration_HF(parameter='C2')
    C3: calibration_HF = calibration_HF(parameter='C3')

class logger_HFV4CH(pydantic.BaseModel):
    description: str = '4 Channel voltage logger'
    sensor_type: typing.Literal['hfv4ch'] = 'hfv4ch'
    logger_model: str = '4 Channel ADS logger'
    logger_configuration: str = 'Raspberry PI board'
    channels: channels_HFV4CH = channels_HFV4CH()
    sn: str = pydantic.Field(default='', description='The serial number and/or MAC address of the sensor')
    comment: str = pydantic.Field(default='')


def convert_HF(datapacket, loggerconfigurations=None):
    """
    Converts raw data into SI units
    """
    funcname = 'convert_HF():'
    logger.debug(funcname)
    try:
        mac = datapacket['sn']
    except:
        mac = None

    #datapacketSI = {}
    devicename = 'DHF_SI_' + mac
    datapacketSI = data_packets.create_datadict(device=devicename, hostinfo=datapacket['_redvypr']['host'], tu=datapacket['t'])


    datapacketSI['type'] = 'HFSI'
    datapacketSI['sensortype'] = datapacket['sensortype']
    datapacketSI['sn'] = datapacket['sn']
    datapacketSI['ts'] = datapacket['ts']
    datapacketSI['np'] = datapacket['np']
    datapacketSI['tUTC'] = datetime.datetime.utcfromtimestamp(datapacket['t']).isoformat(sep=' ',
                                                                                         timespec='milliseconds')
    #print('loggerconfigs',loggerconfigurations)
    if (mac is not None) and (loggerconfigurations is not None):
        flag_mac_calib = False
        try:
            c = loggerconfigurations[mac]
        except:
            c = None

        if c is not None:
            datapacketSI['datatype'] = 'converted_cal'  # Here the id of the calibration could be added
            #print('Found configuration',c)
            if True:
                flag_mac_calib = True
                logger.debug(funcname + 'Found calibration for {:s}, converting'.format(mac))
                for parameter_tuple in c.parameter:
                    parameter = parameter_tuple[0]
                    parameter_calibration = parameter_tuple[1]
                    coeff = parameter_calibration.coeff
                    if parameter == 'HF':
                        data = datapacket['HF']
                        dataSI = data * coeff * 1000
                        datapacketSI['HF'] = dataSI
                    elif 'NTC' in parameter:
                        # Convert data using a Hoge (Steinhardt-Hart) Type of Equation
                        # print('NTC', parameter, coeff['coeff'])
                        if parameter in datapacket.keys():
                            poly = coeff
                            Toff = parameter_calibration.Toff
                            # print('Poly', poly)
                            data = datapacket[parameter]
                            data_tmp = numpy.polyval(poly, numpy.log(data))
                            # print('data_tmp', data_tmp)
                            data_tmp = 1 / data_tmp - Toff
                            dataSI = data_tmp
                            # print('data_SI', dataSI)
                            datapacketSI[parameter] = dataSI

                return datapacketSI

        if flag_mac_calib == False:
            logger.debug(funcname + 'Did not find calibration for {:s}'.format(mac))
            return None

    return None


def parse_HF_raw(data, sensortype='DHFS50', datapacket=None):
    """
    b'$FC0FE7FFFE167225,HF,00012848.3125,6424,-0.001354,848.389,848.389,852.966\n'
    """
    datasplit = data.split(',')
    macstr = datasplit[0][1:17]
    packettype = datasplit[1]  # HF
    if datapacket is None:
        datapacket = {}
    datapacket['type'] = packettype
    datapacket['sn'] = macstr
    datapacket['sensortype'] = sensortype
    valid_package = False
    if len(datasplit) == 8:
        ts = float(datasplit[2])  # Sampling time
        np = int(datasplit[3])  # Number of packet
        datapacket['HF'] = float(datasplit[4])
        datapacket['NTC0'] = float(datasplit[5])
        datapacket['NTC1'] = float(datasplit[6])
        datapacket['NTC2'] = float(datasplit[7])
        datapacket['ts'] = ts
        datapacket['np'] = np
        valid_package = True

    if valid_package:
        return datapacket
    else:
        return None



def parse_HFS_raw(data, sensortype='DHFS50'):
    """
    HFS SI data from the sensor itself
    """
    datasplit = data.split(',')
    macstr = datasplit[0][1:17]
    packettype = datasplit[1]  # HFS
    datapacket = {}
    datapacket['type'] = packettype
    datapacket['sn'] = macstr
    datapacket['sensortype'] = sensortype
    valid_package = False
    if len(datasplit) == 8:
        ts = float(datasplit[2])  # Sampling time
        np = int(datasplit[3])  # Number of packet
        datapacket['HF'] = float(datasplit[4])
        datapacket['NTC0'] = float(datasplit[5])
        datapacket['NTC1'] = float(datasplit[6])
        datapacket['NTC2'] = float(datasplit[7])
        datapacket['ts'] = ts
        datapacket['np'] = np
        valid_package = True

    if valid_package:
        return datapacket
    else:
        return None


def add_keyinfo_HF_raw(datapacket):
    """
    Adds keyinfo to the HF_raw datapacket
    """
    macstr = datapacket['sn']
    # print('Datapacket', datapacket)
    keys = list(datapacket.keys())
    for dkey in keys:
        # print(funcname + ' datakey {:s}'.format(dkey))
        if 'NTC' in dkey:
            unit = 'ohm'
        elif 'HF' in dkey:
            unit = 'V'
        elif dkey == 'T':
            unit = 'CNT'
        else:
            unit = 'NA'

        # print('Datakeyinfo',unit,macstr,dkey)
        data_packets.add_metadata2datapacket(datapacket, datakey=dkey,
                                             unit=unit, description=None,
                                             infokey='sn', info=macstr)
        data_packets.add_metadata2datapacket(datapacket, datakey=dkey,
                                             infokey='sensortype', info='DHFS50')
    return datapacket


def parse_IMU_raw(data, sensortype='DHFS50'):
    """
    b'$FC0FE7FFFE167225,IMU,00012848.3125,6424,-24534,32,-1,-129,-249,0,43872,-7,-1,-75\n'
    """
    datasplit = data.split(',')
    macstr = datasplit[0][1:17]
    packettype = datasplit[1]  # IMU
    datapacket = {}
    datapacket['type'] = packettype
    datapacket['sn'] = macstr
    datapacket['sensortype'] = sensortype
    acc = []
    gyro = []
    mag = []
    T = -9999.9
    valid_package = False
    if len(datasplit) == 14:
        ts = float(datasplit[2])  # Sampling time
        np = int(datasplit[3])  # Number of packet
        # print('{:f} {:d}'.format(ts,np))
        # print('datasplit',datasplit)
        ind_acc = 4
        ind_gyro = 7
        ind_T = 10
        ind_mag = 11
        for n in range(3):
            acc.append(int(datasplit[ind_acc + n]))
        for n in range(3):
            gyro.append(int(datasplit[ind_gyro + n]))
        for n in range(3):
            mag.append(int(datasplit[ind_mag + n]))

        T = int(datasplit[ind_T])
        valid_package = True
        datapacket['acc'] = acc
        datapacket['gyro'] = gyro
        datapacket['mag'] = mag
        datapacket['T'] = T
        datapacket['ts'] = ts
        datapacket['np'] = np

    if valid_package:
        return datapacket
    else:
        return None


def parse_HFV_raw(data, sensortype='PIBOARD'):
    """
    b'$KAL1,HFV,2023-11-03 05:30:29.589,0,-0.021414,1,-0.870808,2,-0.019380,3,-0.000244'
    """
    funcname = __name__ + '.parse_HFV_raw():'
    logger.debug(funcname)

    datasplit = data.split(',')
    snstr = datasplit[0][1:]
    packettype = datasplit[1]  # HFV
    tstr = datasplit[2]
    td = datetime.datetime.strptime(tstr, "%Y-%m-%d %H:%M:%S.%f").replace(tzinfo=zoneinfo.ZoneInfo('UTC'))

    datapacket = {}
    datapacket['type'] = packettype
    datapacket['sn'] = snstr
    datapacket['sensortype'] = sensortype
    datapacket['t'] = td.timestamp()
    for i in range(0, 8, 2):
        ch = datasplit[3 + i]
        voltage = datasplit[3 + i + 1]
        channelname = 'C' + str(ch)
        datapacket[channelname] = float(voltage)

    return datapacket


def process_IMU_packet(dataline,data, macstr, macs_found):
    funcname = __name__ + '.process_IMU_packet()'
    datapacket = parse_IMU_raw(dataline)
    datapacket['t'] = data['t']
    datapacket['devicename'] = 'DHF_raw_' + macstr
    if macstr not in macs_found['IMU']:
        logger.debug(funcname + 'New MAC found, will add datakeyinfo')
        # print('Datapacket', datapacket)
        keys = list(datapacket.keys())
        for dkey in keys:
            # print(funcname + ' datakey {:s}'.format(dkey))
            unit = 'CNT'
            data_packets.add_metadata2datapacket(datapacket, datakey=dkey,
                                                 unit=unit, description=None,
                                                 infokey='sn', info=macstr)
            data_packets.add_metadata2datapacket(datapacket, datakey=dkey,
                                                 infokey='sensortype', info='DHFS50')

        macs_found['IMU'].append(macstr)
    # print('IMU packet',datapacket)


def process_HF_data(dataline, data, device_info, sensorconfig, config, macs_found):
    """
    Processes Heatflow raw data
    """
    funcname = __name__ + '.process_HF_data():'
    datapackets = []
    datapacket = parse_HF_raw(dataline)

    macstr = datapacket['sn']
    sn = datapacket['sn']
    if datapacket is not None:
        devicename = 'DHF_raw_' + macstr
        datapacket['datatype'] = 'raw'
        datapacket['t'] = data['_redvypr']['t']
        datapacket_HF = data_packets.create_datadict(device=devicename, hostinfo=device_info['hostinfo'], tu=data['t'])
        datapacket_HF.update(datapacket)
        if macstr not in macs_found['HF']:
            logger.debug(funcname + 'New MAC found, will add keyinfo')
            datapacket_HF = add_keyinfo_HF_raw(datapacket_HF)

            # print('datapacket',datapacket)
            macs_found['HF'].append(macstr)
        # Convert the data (if calibration found)
        datapackets.append(datapacket_HF)
        if sensorconfig is not None:
            flag_convert = config.sensorconfigurations[sn].convert_rawdata
            if flag_convert:
                datapacket_HFSI = convert_HF(datapacket_HF, config.sensorconfigurations)               # print('datapacket hfsi',datapacket_HFSI)
                # print('sn {:s}', sn)
                if datapacket_HFSI is not None:
                    # Check if units should be set, this is only done once
                    if macstr not in macs_found['HFSI']:
                        logger.debug(funcname + 'New MAC found, will add keyinfo')
                        # print('Datapacket', datapacket)
                        keys = list(datapacket.keys())
                        for dkey in keys:
                            # print(funcname + ' datakey {:s}'.format(dkey))
                            if 'NTC' in dkey:
                                unit = 'degC'
                            elif 'HF' in dkey:
                                unit = 'W m-2'
                            else:
                                unit = 'NA'

                            # print('Datakeyinfo', unit, macstr, dkey)
                            data_packets.add_metadata2datapacket(datapacket_HFSI,
                                                                 datakey=dkey,
                                                                 unit=unit,
                                                                 description=None,
                                                                 infokey='sn',
                                                                 info=macstr)
                            data_packets.add_metadata2datapacket(datapacket_HFSI,
                                                                 datakey=dkey,
                                                                 infokey='sensortype',
                                                                 info='DHFS50')

                        # print('datapacket', datapacket)
                        macs_found['HFSI'].append(macstr)

                datapackets.append(datapacket_HFSI)
        # print('HF packet',datapacket)
        print('DATAPACKETS')
        print('datapackets',datapackets)
        print('DATAPACKETS')
        return datapackets


def process_HFS_data(dataline, data, device_info):
    try:
        datapacket = parse_HFS_raw(dataline)
    except:
        logger.debug(' Could not decode {:s}'.format(str(dataline)), exc_info=True)
        datapacket = None

    if datapacket is not None:
        macstr = datapacket['sn']
        devicename = 'DHF_SI_' + macstr
        datapacket_HFSI = data_packets.create_datadict(device=devicename, hostinfo=device_info['hostinfo'], tu=data['t'])
        datapacket_HFSI['type'] = 'HFSI'
        datapacket_HFSI['datatype'] = 'converted_' + macstr
        datapacket_HFSI.update(datapacket)
        return datapacket_HFSI

        # Averaging the data
        if False:
            dataqueue.put(datapacket_HFS)
            # Do averaging.
            sn = datapacket_HFS['sn']
            try:
                datapacket_HFS_avg = avg_objs[sn].average_data(data)
            except:
                parameters = ['t', 'HF', 'NTC0', 'NTC1', 'NTC2']
                avg_objs[sn] = average_datapacket(datapacket_HFS, parameters, loggerconfig)
                datapacket_HFS_avg = avg_objs[sn].average_data(data)
                logger.debug(funcname + ' could not average:', exc_info=True)

            if datapacket_HFS_avg is not None:
                dataqueue.put(datapacket_HFS_avg)


        return datapacket_HFSI




#
#
# Here the widgets start
#
#
class DHFS50Widget_config(QtWidgets.QWidget):
    """
    Widget to configure a DHFS50 logger
    """

    def __init__(self, *args, sn=None, redvypr_device=None):
        funcname = __name__ + '__init__()'
        super(QtWidgets.QWidget, self).__init__(*args)
        logger.debug(funcname)
        self.device = redvypr_device
        self.nosensorstr = ''
        self.sn = sn
        # Try to find a configuration for the logger
        try:
            self.device.custom_config.sensorconfigurations[self.sn]
            logger.debug(funcname + ' Found a configuration')
        except:
            logger.warning(funcname + ' Did not find a configuration')
            return

        self.layout = QtWidgets.QGridLayout(self)
        self.coeffInput = []
        self.coeff_widget = QtWidgets.QTabWidget()
        print('Configuration', self.device.custom_config.sensorconfigurations[self.sn])
        snlabel = QtWidgets.QLabel('DHFS50 calibration coefficients of\n' + self.sn)
        self.layout.addWidget(snlabel,0,0)

        nparameter = len(list(self.device.custom_config.sensorconfigurations[self.sn].parameter))

        #self.coeff_table = QtWidgets.QTableWidget()
        #self.coeff_table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        self.layout.addWidget(self.coeff_widget, 2, 0)

        self.parameter_buttons = {}
        self.parameter_widgets = {}
        self.parameter_calwidgets = {}
        for i,k_tmp in enumerate(self.device.custom_config.sensorconfigurations[self.sn].parameter):
            k = k_tmp[0]
            calibration = k_tmp[1]
            self.parameter_widgets[k] = QtWidgets.QWidget()
            layout = QtWidgets.QVBoxLayout(self.parameter_widgets[k])
            self.parameter_buttons[k] = QtWidgets.QPushButton('Choose calibration')
            self.parameter_buttons[k].clicked.connect(self.choose_calibration)
            self.parameter_buttons[k].parameter = k
            layout.addWidget(self.parameter_buttons[k])
            calwidget = gui.pydanticQTreeWidget(calibration,dataname = k)
            self.parameter_calwidgets[k] = calwidget
            layout.addWidget(calwidget)
            self.coeff_widget.addTab(self.parameter_widgets[k], k)
            #self.layout.addWidget(self.parameter_buttons[k], 1, i+0)

        #self.fill_coeff_table()

    def choose_calibration(self):
        """
        Let the user choose a calibration for the parameter
        """
        but = self.sender()

        funcname = __name__ + '.choose_calibration():'
        logger.debug(funcname)
        self._sensor_choose_dialog = sensorCoeffWidget(redvypr_device=self.device)
        #self._sensor_choose_dialog.coeff.connect(self.calibration_chosen)
        #self._sensor_choose_dialog.parameter = self.sender().parameter
        self._sensor_choose_dialog.show()

    def calibration_chosen(self, calibration_sent):
        funcname = __name__ + '.calibration_chosen():'
        logger.debug(funcname)
        #print('Coeff sent',calibration_sent)
        calibration_new = calibration_sent['calibration']
        # Use only the chosen parameter, remove the rest
        parameter = self.sender().parameter
        # print('Coeff',coeff,'sn',sn,'channel',channel)
        ## Make a test configuration for KAL1
        # channelconfig = redvypr_config.configuration(calibration_template_channel)
        # channelconfig['sn'] = self.nosensorstr  # The serialnumber of the channel
        if True:
            logger.debug('Adding {} to channel {:s}'.format(calibration_sent, parameter))
            #channels = self.device.config.sensorconfigurations[self.sn].channels
            #setattr(channels, channelname, calibration)
            calibration_old = self.device.custom_config.sensorconfigurations[self.sn].parameter
            print('calibration_old', type(calibration_old))
            print('calibration_new', type(calibration_sent))
            print('hallohallo',calibration_old.model_validate(calibration_new))
            # Check if the new coefficients fit
            try:
                calibration_old.model_validate(calibration_new)
                setattr(self.device.custom_config.sensorconfigurations[self.sn].parameter, parameter, calibration_new)
                self.parameter_calwidgets[parameter].reload_data(calibration_new)
            except:
                logger.debug('Calibration format does not fit',exc_info = True)



        #self.fill_coeff_table()
        print('configuration', self.device.custom_config.sensorconfigurations[self.sn])


#
#
#
#
#
class HFVWidget_config(QtWidgets.QWidget):
    """
    Widget to display information from a 4 Channel Voltage board
    """

    def __init__(self, *args, sn=None, redvypr_device=None):
        funcname = __name__ + '__init__()'
        super(QtWidgets.QWidget, self).__init__(*args)
        logger.debug(funcname)
        self.device = redvypr_device
        self.nosensorstr = ''
        self.sn = sn

        # Get the channel names
        ch_tmp = channels_HFV4CH()
        self.channelnames = []
        for i, ch in enumerate(ch_tmp):
            print(ch)
            self.channelnames.append(ch[0])

        print('Channelsnames',self.channelnames)

        # Try to find a configuration for the logger
        try:
            self.device.custom_config.sensorconfigurations[self.sn]
            logger.debug(funcname + ' Found a configuration')
        except:
            logger.warning(funcname + ' Did not find a configuration')
            return

        self.layout = QtWidgets.QGridLayout(self)
        self.coeffInput = []
        self.coeff_widget = QtWidgets.QWidget()
        self.create_coeff_widget()
        self.layout.addWidget(self.coeff_widget, 0, 0, 1, 2)


    def create_coeff_widget(self):
        """

        """
        # Coeffs
        # Serialnum
        # Sensortype
        # Comment
        # Show converted data

        self.coeff_layout = QtWidgets.QGridLayout(self.coeff_widget)
        self.use_coeff_checks = []
        self.serialnum_texts = []
        self.comment_texts = []
        self.sensor_choose_buttons = []

        # Get the channel names
        for i,chname in enumerate(self.channelnames):
            but = QtWidgets.QPushButton('Sensor {:s}'.format(chname))
            but.clicked.connect(self.choose_sensor)
            self.sensor_choose_buttons.append(but)
            self.coeff_layout.addWidget(but,0,i)
            but.channelname  = chname
            but.channelindex = i

        row = 0
        col = 0
        self.coeff_table = QtWidgets.QTableWidget()
        self.coeff_layout.addWidget(self.coeff_table, 1, 0,1,4)

        self.index_serialnum = 0
        self.index_coeff = 2
        self.index_model = 1
        self.index_date = 3
        self.index_fname = 4

        # Fill table with configuration data
        #addresses = ["<broadcast>", "<IP>", myip]
        #completer = QtWidgets.QCompleter(addresses)
        #self.addressline.setCompleter(completer)
        self.fill_coeff_table()

    def sensor_chosen(self,coeff_sent):
        funcname = __name__ + '.sensor_chosen():'
        logger.debug(funcname)

        calibration = coeff_sent['calibration']#
        sn = calibration.sn
        # Use only the chosen parameter, remove the rest
        channelname = self.sender().channelname
        #print('Coeff',coeff,'sn',sn,'channel',channel)
        ## Make a test configuration for KAL1
        #channelconfig = redvypr_config.configuration(calibration_template_channel)
        #channelconfig['sn'] = self.nosensorstr  # The serialnumber of the channel
        if sn == self.nosensorstr:
            logger.debug(funcname + 'Removing sensor')
        else:
            logger.debug('Adding {:s} to channel {:s}'.format(sn, channelname))
            channels = self.device.custom_config.sensorconfigurations[self.sn].channels
            setattr(channels,channelname,calibration)

        self.fill_coeff_table()
        #print('configuration',self.device.config['sensorconfigurations'][self.sn])

    def choose_sensor(self):
        """
        Let the user choose a sensor for the channel
        """
        but = self.sender()

        funcname = __name__ + '.choose_sensor():'
        logger.debug(funcname)
        self._sensor_choose_dialog = sensorCoeffWidget(redvypr_device=self.device)
        self._sensor_choose_dialog.coeff.connect(self.sensor_chosen)
        self._sensor_choose_dialog.channelname = self.sender().channelname
        self._sensor_choose_dialog.show()


    def fill_coeff_table(self):
        """

        """
        funcname = __name__ + '.fill_coeff_table()'
        logger.debug(funcname)


        self.coeff_table.clear()
        self.coeff_table.setRowCount(5)
        self.coeff_table.setColumnCount(len(self.channelnames))
        self.coeff_table.setVerticalHeaderLabels(
            ['Serialnum', 'Model', 'Coefficient', 'Calibration date', 'Comment'])
        self.coeff_table.setHorizontalHeaderLabels(self.channelnames )
        for ch,ch_tmp in enumerate(self.device.custom_config.sensorconfigurations[self.sn].channels):
            chstr = ch_tmp[0]
            sn_sensor_attached = None
            channels = self.device.custom_config.sensorconfigurations[self.sn].channels
            calibration = getattr(channels,chstr)
            print('Got channel',calibration)
            sn_sensor_attached = calibration.sn

            if sn_sensor_attached is None:
                logger.warning('Could not find coeffcients for {:s}'.format(sn_sensor_attached))
                continue
            elif len(sn_sensor_attached) == 0:
                logger.warning('No sensor attached at channel {:s}'.format(chstr))
                continue
            else:
                print('Using calibration',calibration)
                # Try to find coefficients for the sensor
                try:
                    coeff = str(calibration.coeff)
                except Exception as e:
                    coeff = numpy.NaN
                    logger.debug(funcname)
                    logger.exception(e)
                    logger.debug(funcname)

                #
                try:
                    fname = str(calibration['original_file'])
                except Exception as e:
                    fname = ''
                    logger.debug(funcname, exc_info=True)

                # Try to find coefficients for the sensor
                try:
                    coeff_date = str(calibration.date)
                except Exception as e:
                    coeff_date = ''
                    logger.debug(funcname)
                    logger.exception(e)
                    logger.debug(funcname)

                # Try to find coefficients for the sensor
                try:
                    model = str(calibration.sensor_model)
                except Exception as e:
                    model = ''
                    logger.debug(funcname)
                    logger.exception(e)
                    logger.debug(funcname)

                index_col = ch
                index_row = self.index_serialnum
                item = QtWidgets.QTableWidgetItem(str(sn_sensor_attached))
                self.coeff_table.setItem(index_row,index_col,item)

                if coeff is numpy.NaN:
                    coeffstr = ''
                else:
                    coeffstr = str(coeff)

                index_row = self.index_coeff
                item = QtWidgets.QTableWidgetItem(coeffstr)
                self.coeff_table.setItem(index_row, index_col, item)

                index_row = self.index_model
                item = QtWidgets.QTableWidgetItem(str(model))
                self.coeff_table.setItem(index_row, index_col, item)

                index_row = self.index_date
                item = QtWidgets.QTableWidgetItem(str(coeff_date))
                self.coeff_table.setItem(index_row, index_col, item)

                #index_row = self.index_fname
                #item = QtWidgets.QTableWidgetItem(str(fname))
                #self.coeff_table.setItem(index_row, index_col, item)

        self.coeff_table.resizeColumnsToContents()

class HFVWidget(QtWidgets.QWidget):
    """
    Widget to display information from a 4 Channel Voltage board
    """

    def __init__(self, *args, sn=None, redvypr_device=None):
        funcname = __name__ + '__init__()'
        super(QtWidgets.QWidget, self).__init__(*args)
        logger.debug(funcname)
        self.device = redvypr_device

        self.sn = sn
        # Try to find a configuration for the logger
        try:
            self.device.custom_config.sensorconfigurations[self.sn]
            logger.debug(funcname + ' Found a configuration')
        except:
            logger.warning(funcname + ' Did not find a configuration, will need to create one')
            return
            # Make a test configuration for KAL1
            #channelconfig = config.configuration(calibration_template_channel)

        self.rawconvtabs = QtWidgets.QTabWidget()

        # Add table with the data (raw data)
        self.datatable_raw = QtWidgets.QTableWidget()
        self.datatable_raw.setColumnCount(6)
        self.datatable_raw.setRowCount(2)
        self.datatable_raw.resizeRowsToContents()
        self.datatable_raw.horizontalHeader().hide()
        self.datatable_raw.verticalHeader().hide()
        self.datatable_raw.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        self.datatable_raw.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)

        # Add table with the data (converted data)
        self.datatable_conv = QtWidgets.QTableWidget()
        self.datatable_conv.setColumnCount(6)
        self.datatable_conv.setRowCount(2)
        self.datatable_conv.horizontalHeader().hide()
        self.datatable_conv.verticalHeader().hide()
        self.datatable_conv.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        self.datatable_conv.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)

        # Fill the tables
        for i in range(4):
            item = QtWidgets.QTableWidgetItem('CH' + str(i))
            self.datatable_raw.setItem(0, i + 1, item)
            item = QtWidgets.QTableWidgetItem('CH' + str(i))
            self.datatable_conv.setItem(0, i + 1, item)

        item = QtWidgets.QTableWidgetItem('Time')
        self.datatable_raw.setItem(0, 0, item)
        item = QtWidgets.QTableWidgetItem('Time')
        self.datatable_conv.setItem(0, 0, item)
        item = QtWidgets.QTableWidgetItem('Unit')
        self.datatable_raw.setItem(0, 5, item)
        item = QtWidgets.QTableWidgetItem('Unit')
        self.datatable_conv.setItem(0, 5, item)
        item = QtWidgets.QTableWidgetItem('V')
        self.datatable_raw.setItem(1, 5, item)
        item = QtWidgets.QTableWidgetItem('Wm-2')
        self.datatable_conv.setItem(1, 5, item)

        self.layout = QtWidgets.QGridLayout(self)
        self.layout.addWidget(self.rawconvtabs,0,0)

        self.rawdatawidget = QtWidgets.QWidget()
        rawlayout = QtWidgets.QGridLayout(self.rawdatawidget)
        rawlayout.addWidget(self.datatable_raw, 0, 0, 1, 2)

        self.convdatawidget = QtWidgets.QWidget()
        convlayout = QtWidgets.QGridLayout(self.convdatawidget)
        convlayout.addWidget(self.datatable_conv, 0, 0, 1, 2)


        self.plot_widgets_HFV = []
        self.plot_widgets_HFVSI = []
        self.coeffInput = []
        self.layout.setRowStretch(0, 1)
        row = 0
        col = 0
        for i in range(4):
            # Create plots
            plot_widgets.logger.setLevel(logging.INFO)
            config = XYplotWidget.configXYplot(title='Voltage CH{:d}'.format(i))
            plot_widget_HFV = XYplotWidget.XYPlotWidget(config=config, redvypr_device=self.device.redvypr)
            # Subscribe to channel
            plot_widget_HFV.set_line(0, 'C{:d}'.format(i) + '/{HFV_raw_.*}', name='HFV C{:d}'.format(i), color='red', linewidth=2)
            self.plot_widgets_HFV.append(plot_widget_HFV)
            rawlayout.addWidget(plot_widget_HFV, row+1, col)
            rawlayout.setRowStretch(row+1, 3)

            config = XYplotWidget.configXYplot(title='Heat flow CH{:d}'.format(i))
            plot_widget_HFVSI = XYplotWidget.XYPlotWidget(config=config, redvypr_device=self.device.redvypr)
            # Subscribe to channel
            plot_widget_HFVSI.set_line(0, 'C{:d}'.format(i) + '/{HFV_SI_.*}', name='HFVSI C{:d}'.format(i), color='red',
                                     linewidth=2)
            self.plot_widgets_HFVSI.append(plot_widget_HFVSI)
            convlayout.addWidget(plot_widget_HFVSI, row + 1, col)
            convlayout.setRowStretch(row + 1, 1)

            col += 1
            if col > 1:
                col = 0
                row += 1

        self.rawconvtabs.addTab(self.rawdatawidget,'Raw data')
        self.rawconvtabs.addTab(self.convdatawidget, 'Converted data')


    def update_data(self, data):
        """
        Updating the local data with datapacket
        """
        funcname = __name__ + '.update_data()'
        logger.debug(funcname)
        mac = data['sn']
        if mac in self.sn:
            #print(funcname + ' Got data',data)
            # Update table
            if data['type'] == 'HFV':  # Raw data
                tdata = datetime.datetime.utcfromtimestamp(data['t'])
                tdatastr = tdata.strftime('%Y-%m-%d %H:%M:%S.%f')
                item = QtWidgets.QTableWidgetItem(tdatastr)
                self.datatable_raw.setItem(1,0, item)
                for chi in range(4):
                    channelname = 'C' + str(chi)
                    chdata = "{:+.6f}".format(data[channelname])
                    item = QtWidgets.QTableWidgetItem(str(chdata))
                    self.datatable_raw.setItem(1, chi + 1, item)

                #self.datatable_raw.resizeColumnsToContents()
                for i in range(4):
                    try:
                        #print('data',data,i)
                        self.plot_widgets_HFV[i].update_plot(data)
                    except Exception as e:
                        pass
            elif data['type'] == 'HFVSI':  # Converted data
                try:
                    tdata = datetime.datetime.utcfromtimestamp(data['t'])
                    tdatastr = tdata.strftime('%Y-%m-%d %H:%M:%S.%f')
                    item = QtWidgets.QTableWidgetItem(tdatastr)
                    self.datatable_conv.setItem(1, 0, item)
                    for chi in range(4):
                        try:
                            channelname = 'C' + str(chi)
                            chdata = "{:+.6f}".format(data[channelname])
                            item = QtWidgets.QTableWidgetItem(str(chdata))
                            self.datatable_conv.setItem(1, chi + 1, item)
                        except:
                            pass
                except:
                    pass

                for i in range(4):
                    print('Updating')
                    try:
                        self.plot_widgets_HFVSI[i].update_plot(data)
                    except Exception as e:
                        logger.info('Error',exc_info=True)
                        pass



class DHFSWidget(QtWidgets.QWidget):
    """
    Widget to display Digital Heat flow sensor data
    """
    def __init__(self,*args, sn=None, redvypr_device=None):
        funcname = __name__ + '__init__()'
        logger.debug(funcname)
        super(QtWidgets.QWidget, self).__init__(*args)

        self.device = redvypr_device
        self.sn = sn
        self.configuration = None

        if self.device is not None:
            try:
                self.configuration = self.device.custom_config.sensorconfigurations[sn]
            except Exception as e:
                logger.exception(e)

        #self.parameter = ['HF', 'NTC0', 'NTC1', 'NTC2']
        self.parameter = []
        self.parameter_unit = []
        self.parameter_unitconv = []
        print('Adding parameter')
        for ptmp in self.configuration.parameter:
            #print('Parameter',ptmp)
            p = ptmp[0]
            self.parameter.append(p)
            self.parameter_unit.append(None)
            self.parameter_unitconv.append(None)

        #self.create_coeff_widget()
        self.XYplots = []
        self.layout = QtWidgets.QVBoxLayout(self)
        self.hlayout = QtWidgets.QHBoxLayout(self)
        self.configButton = QtWidgets.QPushButton('Configure')
        self.configButton.clicked.connect(self.__configClicked__)
        # Add a widget for the raw data
        self.rawdataWidget = QtWidgets.QWidget()
        self.rawdataWidget_layout = QtWidgets.QVBoxLayout(self.rawdataWidget)
        label = QtWidgets.QPushButton('Raw data')
        label.setEnabled(False)
        self.rawdataWidget_layout.addWidget(label)
        # Add a table for the raw data
        self.datatable_raw = QtWidgets.QTableWidget()
        self.rawdataWidget_layout.addWidget(self.datatable_raw)
        # number of columns
        ncols = 4
        self.datatable_raw.setColumnCount(ncols)
        self.datatable_raw_parameterheader = ['Time']
        for i, p in enumerate(self.parameter):
            self.datatable_raw_parameterheader.append(p)

        self.datatable_raw.setRowCount(len(self.datatable_raw_parameterheader))
        self.datatable_raw.setVerticalHeaderLabels(self.datatable_raw_parameterheader)
        self.datatable_raw.setHorizontalHeaderLabels(['Parameter','Unit','Data'])
        self.datatable_raw.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        self.datatable_raw.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)

        # Converted data
        self.convdataWidget = QtWidgets.QWidget()
        self.convdataWidget_layout = QtWidgets.QVBoxLayout(self.convdataWidget)
        label = QtWidgets.QPushButton('Converted data')
        label.setEnabled(False)
        self.convdataWidget_layout.addWidget(label)
        # Add a table for the conv data
        self.datatable_conv = QtWidgets.QTableWidget()
        self.convdataWidget_layout.addWidget(self.datatable_conv)
        self.datatable_conv.setColumnCount(ncols)
        self.datatable_conv.setRowCount(len(self.datatable_raw_parameterheader))
        self.datatable_conv.setVerticalHeaderLabels(self.datatable_raw_parameterheader)
        self.datatable_conv.setHorizontalHeaderLabels(['Parameter', 'Unit', 'Data'])
        self.datatable_conv.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        self.datatable_conv.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)


        if self.configuration.show_rawdata:
            self.hlayout.addWidget(self.rawdataWidget)
        if self.configuration.show_convdata:
            self.hlayout.addWidget(self.convdataWidget)

        self.layout.addLayout(self.hlayout)
        self.layout.addWidget(self.configButton)
        self.add_XYplots()


    def __configClicked__(self):

        #self.configWidget = DHFS50Widget_config(sn = self.sn, redvypr_device=self.device)
        self.configWidget = sensorConfigWidget(sensor=self.configuration, calibrations=self.device.custom_config.calibrations)
        self.configWidget.config_changed_flag.connect(self.__configChanged__)
        self.configWidget.show()

    def __configChanged__(self):
        print('Config changed')
        # Updating the widgets

    def add_XYplots(self):

        self.device_subscriptions_raw = []
        self.device_subscriptions_si = []
        for p_index,p in enumerate(self.parameter):
            # The parameter
            # Raw
            subtmp_raw = '/d:{{DHF_raw_{}}}/k:{}/'.format(self.sn, p)
            self.device_subscriptions_raw.append(subtmp_raw)
            config = XYplotWidget.configXYplot(title='{} {}'.format(self.sn, p))
            self.plot_widget_HF_raw = XYplotWidget.XYPlotWidget(config=config, redvypr_device=self.device)
            self.plot_widget_HF_raw.set_line(0, subtmp_raw, name=p, color='red', linewidth=2)
            # Create the button to plot the widget
            but = QtWidgets.QPushButton('Plot')
            but.clicked.connect(self.__plot_clicked__)
            self.datatable_raw.setCellWidget(p_index + 1, 3, but)
            but.__widget__ = self.plot_widget_HF_raw
            # Converted
            subtmp_si = '/d:{{DHF_SI_{}}}/k:{}/'.format(self.sn, p)
            self.device_subscriptions_si.append(subtmp_si)
            config = XYplotWidget.configXYplot(title='{} (converted)'.format(p))
            self.plot_widget_HF_SI = XYplotWidget.XYPlotWidget(config=config, redvypr_device=self.device)
            self.plot_widget_HF_SI.set_line(0, subtmp_si, name=p, color='red', linewidth=2)
            # Create the button to plot the widget
            but = QtWidgets.QPushButton('Plot')
            but.clicked.connect(self.__plot_clicked__)
            self.datatable_conv.setCellWidget(p_index + 1, 3, but)
            but.__widget__ = self.plot_widget_HF_SI
            # Add all the plots, such that they can be updated
            self.XYplots.append(self.plot_widget_HF_raw)
            self.XYplots.append(self.plot_widget_HF_SI)

            if False: # Average data
                print('Average data!! 0')
                self.avgWidget = QtWidgets.QWidget()
                self.avgWidget_layout = QtWidgets.QVBoxLayout(self.avgWidget)
                self.rawconvtabs.addTab(self.avgWidget, 'Average')

                config = XYplotWidget.configXYplot(title='Heat flow average')
                self.plot_widget_HF_SI_AVG = XYplotWidget.XYPlotWidget(config=config, redvypr_device=redvypr_device, add_line=False)
                for i,avg_s in enumerate(self.configuration.avg_data):
                    avg_str = str(avg_s) + 's'
                    avg_datakey = 'HF_avg_{}'.format(avg_str)
                    std_datakey = 'HF_std_{}'.format(avg_str)
                    raddress = RedvyprAddress(datakey=avg_datakey, devicename='{DHF_SIAVG_.*}')
                    raddress_std = RedvyprAddress(datakey=std_datakey, devicename='{DHF_SIAVG_.*}')
                    address = raddress.address_str
                    address_std = raddress_std.address_str
                    #'HF_avg_10s'
                    #Publishing avg:10s packet
                    #{'_redvypr': {'t': 1707832080.2044983, 'device': 'DHF_SIAVG_FC0FE7FFFE164EE0', 'host': {'hostname': 'redvypr', 'tstart': 1708066243.1493528, 'addr': '192.168.178.157', 'uuid': '176473386979093-187'}}, 't_avg_10s': 1707832073.2196076, 't_std_10s': 3.4081697616298503, 't_n_10s': 6, 'HF_avg_10s': 3.607291333333333, 'HF_std_10s': 0.04859570904468358, 'HF_n_10s': 6, 'NTC0_avg_10s': 21.016166666666663, 'NTC0_std_10s': 0.0003726779962504204, 'NTC0_n_10s': 6, 'NTC1_avg_10s': 21.119, 'NTC1_std_10s': 0.0, 'NTC1_n_10s': 6, 'NTC2_avg_10s': 21.113333333333333, 'NTC2_std_10s': 0.00047140452079160784, 'NTC2_n_10s': 6}
                    self.plot_widget_HF_SI_AVG.add_line(address, name=address, linewidth=1, error_addr=address_std)
                    self.avgWidget_layout.addWidget(self.plot_widget_HF_SI_AVG)

                # Add the three temperature sensors as well
                for NNTC in range(0,3):
                    config = XYplotWidget.configXYplot(title='Temperature (average)')
                    plot_widget_NTC_SI_AVG = XYplotWidget.XYPlotWidget(config=config, redvypr_device=redvypr_device,
                                                                       add_line=False)
                    for i, avg_s in enumerate(self.configuration.avg_data):
                        avg_str = str(avg_s) + 's'
                        avg_datakey = 'NTC{}_avg_{}'.format(NNTC,avg_str)
                        std_datakey = 'NTC{}_std_{}'.format(NNTC,avg_str)
                        raddress = RedvyprAddress(datakey=avg_datakey, devicename='{DHF_SIAVG_.*}')
                        raddress_std = RedvyprAddress(datakey=std_datakey, devicename='{DHF_SIAVG_.*}')
                        address = raddress.address_str
                        address_std = raddress_std.address_str
                        plot_widget_NTC_SI_AVG.add_line(address, name=address, linewidth=1, error_addr=address_std)
                        self.avgWidget_layout.addWidget(plot_widget_NTC_SI_AVG)
                        self.XYplots.append(plot_widget_NTC_SI_AVG)

            if False:
                # Add all the plots, such that they can be updated
                self.XYplots.append(self.plot_widget_HF_SI_AVG)
                self.XYplots.append(self.plot_widget_HF_raw)
                self.XYplots.append(self.plot_widget_NTC_raw)
                self.XYplots.append(self.plot_widget_HF_SI)
                self.XYplots.append(self.plot_widget_NTC_SI)


                self.layout_raw.addWidget(self.plot_widget_HF_raw,1,0)
                self.layout_raw.addWidget(self.plot_widget_NTC_raw,2,0)
                self.layout_conv.addWidget(self.plot_widget_HF_SI,1,0)
                self.layout_conv.addWidget(self.plot_widget_NTC_SI,2,0)

    def __plot_clicked__(self):
        but = self.sender()
        but.__widget__.show()

    def update_data(self, data):
        """
        Updating the local data with datapacket
        """
        funcname = __name__ + '__update__()'
        logger.debug(funcname)
        #print('Got data for widget',data)
        mac = data['sn']

        col_parameter = 0
        col_unit = 1
        col_data = 2
        row_time = 0
        #print('Got data', data['_redvypr']['device'],mac)
        for plot in self.XYplots:
            print('plot update',len(plot.config.lines), mac)
            plot.update_plot(data)

        if mac in self.sn:
            # Update table
            if data['type'] == 'HF':  # Raw data
                #print('Updating')
                # Time
                tdata = datetime.datetime.utcfromtimestamp(data['t'])
                tdatastr = tdata.strftime('%Y-%m-%d %H:%M:%S.%f')
                item = QtWidgets.QTableWidgetItem(tdatastr)
                self.datatable_raw.setItem(row_time, col_data, item)
                for p_index,p in enumerate(self.parameter):
                    # The parameter
                    item = QtWidgets.QTableWidgetItem(p)
                    self.datatable_raw.setItem(p_index + 1, col_parameter, item)
                    # The data
                    try:
                        chdata = "{:+.6f}".format(data[p])
                        item = QtWidgets.QTableWidgetItem(str(chdata))
                        self.datatable_raw.setItem(p_index + 1, col_data, item)
                    except Exception as e:
                        logger.debug(funcname, exc_info=True)

                    if self.parameter_unit[p_index] is None:
                        #datastream = data_packets.get_datastream_from_data(data,p)
                        datastream = RedvyprAddress(data, datakey = p)
                        #print('datastream',datastream)
                        #keyinfo = self.device.get_metadata_datakey(datastream)
                        keyinfo = {}
                        try:
                            self.parameter_unit[p_index] = str(keyinfo['unit'])
                        except:
                            self.parameter_unit[p_index] = str('NA')

                        item = QtWidgets.QTableWidgetItem(self.parameter_unit[p_index])
                        self.datatable_raw.setItem(p_index + 1, col_unit, item)
                        #print('Keyinfo', keyinfo)
                        #print('Parameter', p)
                        #print('datastream',datastream)
                        #print('keyinfo', keyinfo)

            elif data['type'] == 'HFSI':  # converted data
                print('Updating HFSI')
                # Time
                tdata = datetime.datetime.utcfromtimestamp(data['t'])
                tdatastr = tdata.strftime('%Y-%m-%d %H:%M:%S.%f')
                item = QtWidgets.QTableWidgetItem(tdatastr)
                self.datatable_conv.setItem(row_time, col_data, item)
                for p_index, p in enumerate(self.parameter):
                    print('parameter',p)
                    # The parameter
                    item = QtWidgets.QTableWidgetItem(p)
                    self.datatable_conv.setItem(p_index + 1, col_parameter, item)
                    if self.parameter_unitconv[p_index] is None:
                        datastream = RedvyprAddress(data, datakey=p)
                        ##datastream = data_packets.get_datastream_from_data(data, p)
                        #keyinfo = self.device.get_metadata_datakey(datastream)
                        keyinfo = {}
                        try:
                            self.parameter_unitconv[p_index] = str(keyinfo['unit'])
                        except:
                            self.parameter_unitconv[p_index] = str('NA')

                        item = QtWidgets.QTableWidgetItem(self.parameter_unitconv[p_index])
                        self.datatable_conv.setItem(p_index + 1, col_unit, item)
                        # print('Keyinfo', keyinfo)
                        #print('Parameter', p)
                        #print('datastream', datastream)
                        #print('keyinfo', keyinfo)

                    # Update the data
                    try:
                        chdata = "{:+.6f}".format(data[p])
                        item = QtWidgets.QTableWidgetItem(str(chdata))
                        self.datatable_conv.setItem(p_index + 1, col_data, item)
                    except Exception as e:
                        logger.debug(funcname, exc_info=True)

        # Fill the datatables









import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import logging
import sys
import yaml
import copy
import os
from redvypr.device import RedvyprDevice
from redvypr.data_packets import check_for_command
from redvypr.packet_statistic import do_data_statistics, create_data_statistic_dict

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('fileread')
logger.setLevel(logging.DEBUG)

description = "Reads a file and publishes the data as chunks defined by the users"
config_template = {}
config_template['name']              = 'fileread'
config_template['files']             = {'type':'list','description':'List of files ro replay'} # TODO, list
config_template['loop']              = {'type':'bool','default':True,'description':'Loop over all files if set'}
config_template['speedup']           = {'type':'float','description':'Speedup factor of the data'}
config_template['redvypr_device']    = {}
config_template['redvypr_device']['publishes']   = True
config_template['redvypr_device']['subscribes']  = False
config_template['redvypr_device']['description'] = description
redvypr_devicemodule = True
def get_packets(filestream=None):
    funcname = __name__ + '.get_packets()'
    packets = []
    if(filestream is not None):
        data = filestream.read()
        for databs in data.split('...\n'): # Split the text into single subpackets
            try:
                data = yaml.safe_load(databs)
                if(data is not None):
                    packets.append(data)
                    
            except Exception as e:
                logger.debug(funcname + ': Could not decode message {:s}'.format(str(databs)))
                data = None
    
        return packets
    else:
        return None
        
    


def start(device_info, config={'filename': ''}, dataqueue=None, datainqueue=None, statusqueue=None):
    funcname = __name__ + '.start()'
    logger.debug(funcname + ':Opening reading:')
    files = config['files']
    t_sent = 0 # The time the last packets was sent
    t_packet_old = 1e12 # The time the last packet had (internally)
    #
    try:
        config['speedup']
    except:
        config['speedup'] = 1.0 # Realtime
    
    speedup = config['speedup']    
    #
    try:
        config['loop']
    except:
        config['loop'] = False
        
    loop = config['loop']
    
    
    statistics = create_data_statistic_dict()
    
    bytes_read         = 0
    packets_read       = 0
    bytes_read_total   = 0
    packets_read_total = 0
    
    tfile           = time.time() # Save the time the file was created
    tflush          = time.time() # Save the time the file was created
    f = None    
    nfile = 0
    while True:
        tcheck      = time.time()
        try:
            data = datainqueue.get(block=False)
        except:
            data = None
        if (data is not None):
            command = check_for_command(data, thread_uuid=device_info['thread_uuid'])
            # logger.debug('Got a command: {:s}'.format(str(data)))
            if (command is not None):
                sstr = funcname + ': Command is for me: {:s}'.format(str(command))
                logger.debug(sstr)
                if command == 'stop':
                    logger.debug('Stopping')
                    try:
                        statusqueue.put_nowait(sstr)
                    except:
                        pass
                    break

        packets = get_packets(f)
        if(packets is None):
            FLAG_NEW_FILE = True
        else:
            FLAG_NEW_FILE = False
            
        if(packets is not None):
            for p in packets:
                t_now = time.time()
                dt = t_now - t_sent
                dt_packet = (p['_redvypr']['t'] - t_packet_old)/speedup
                if(dt_packet < 0):
                    dt_packet = 0

                if True:
                    time.sleep(dt_packet)
                    t_sent = time.time()
                    t_packet_old = p['_redvypr']['t']
                    dataqueue.put(p)

                sstr = 'Sending packet in {:f} s.'.format(dt_packet)
                logger.debug(sstr)
                try:
                    statusqueue.put_nowait(sstr)
                except:
                    pass


            FLAG_NEW_FILE = True
            f.close()
            
        if(FLAG_NEW_FILE):
            if(nfile >= len(files)):
                if(loop == False):
                    sstr = funcname + ': All files read, stopping now.'
                    try:
                        statusqueue.put_nowait(sstr)
                    except:
                        pass
                    logger.info(sstr)
                    break
                else:
                    sstr = funcname + ': All files read, loop again.'
                    try:
                        statusqueue.put_nowait(sstr)
                    except:
                        pass

                    logger.info(sstr)
                    nfile = 0
            
            filename = files[nfile]
            sstr = 'Opening file {:s}'.format(filename)
            try:
                statusqueue.put_nowait(sstr)
            except:
                pass
            logger.info(sstr)
            f = open(filename)
            nfile += 1
        
        time.sleep(0.05)


#
#
# The init widget
#
#
class initDeviceWidget(QtWidgets.QWidget):
    connect      = QtCore.pyqtSignal(RedvyprDevice) # Signal requesting a connect of the datainqueue with available dataoutqueues of other devices
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout        = QtWidgets.QGridLayout(self)
        self.file_statistics = {}
        self.device   = device
        self.label    = QtWidgets.QLabel("rawdatareplay setup")
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setStyleSheet(''' font-size: 24px; font: bold''')
        self.config_widgets= [] # A list of all widgets that can only be used of the device is not started yet
        # Input output widget
        self.inlabel  = QtWidgets.QLabel("Filenames") 
        self.inlabel.setStyleSheet(''' font-size: 20px; font: bold''')
        self.inlist   = QtWidgets.QTableWidget()
        self.inlist.setColumnCount(4)
        self.inlist.setRowCount(1)
        self.inlist.setSortingEnabled(True)
        
        self.__filelistheader__ = ['Name','Packets','First date','Last date']
        self.inlist.setHorizontalHeaderLabels(self.__filelistheader__)
        self.addfilesbtn   = QtWidgets.QPushButton("Add files")
        self.addfilesbtn.clicked.connect(self.add_files)
        self.remfilesbtn   = QtWidgets.QPushButton("Rem files")
        self.remfilesbtn.clicked.connect(self.rem_files)
        self.scanfilesbtn   = QtWidgets.QPushButton("Scan files")
        self.scanfilesbtn.clicked.connect(self.scan_files_clicked)
        self.config_widgets.append(self.inlist)
        self.config_widgets.append(self.addfilesbtn)
        # Looping the data?
        self.loop_checkbox = QtWidgets.QCheckBox('Loop')
        # Speedup
        self.speedup_edit  = QtWidgets.QLineEdit(self)
        onlyDouble = QtGui.QDoubleValidator()
        self.speedup_edit.setValidator(onlyDouble)
        self.speedup_edit.setToolTip('Speedup of the packet replay.')
        self.speedup_label = QtWidgets.QLabel("Speedup factor")
        self.speedup_edit.setText("1.0")
        
        
        self.startbtn = QtWidgets.QPushButton("Start replay")
        self.startbtn.clicked.connect(self.start_clicked)
        self.startbtn.setCheckable(True)
        self.startbtn.setSizePolicy(QtWidgets.QSizePolicy.Preferred,QtWidgets.QSizePolicy.Expanding)
        
        layout.addWidget(self.label,0,0,1,-1)
        
        layout.addWidget(self.addfilesbtn,1,0,1,-1)               
        layout.addWidget(self.remfilesbtn,2,0,1,-1)          
        layout.addWidget(self.scanfilesbtn,3,0,1,-1)
        layout.addWidget(self.inlabel,4,0,1,-1,QtCore.Qt.AlignCenter)         
        layout.addWidget(self.inlist,5,0,1,-1)
        layout.addWidget(self.loop_checkbox,6,0)
        layout.addWidget(self.speedup_label,6,1,1,1,QtCore.Qt.AlignRight)
        layout.addWidget(self.speedup_edit,6,2,1,1,QtCore.Qt.AlignRight)
        layout.addWidget(self.startbtn,7,0,2,-1)

        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.update_buttons)
        self.statustimer.start(500)

    def inspect_data(self, filename, rescan=False):
        """ Inspects the files for possible datastreams in the file with the filename located in config['files'][fileindex].
        """
        funcname = self.__class__.__name__ + '.inspect_data()'
        logger.debug(funcname)
        try:
            stat = self.file_statistics[filename]
            FLAG_HASSTAT = True
        except:
            FLAG_HASSTAT = False

        if (rescan or (FLAG_HASSTAT == False)):
            logger.debug(funcname + ': Scanning file {:s}'.format(filename))
            stat = create_data_statistic_dict()
            # Create tmin/tmax in statistics
            stat['t_min'] = 1e12
            stat['t_max'] = -1e12
            f = open(filename)
            packets = get_packets(f)
            f.close()
            for p in packets:
                stat = do_data_statistics(p, stat)
                tminlist = [stat['t_min'], p['_redvypr']['t']]
                tmaxlist = [stat['t_max'], p['_redvypr']['t']]
                stat['t_min'] = min(tminlist)
                stat['t_max'] = max(tmaxlist)

            self.file_statistics[filename] = stat
        else:
            logger.debug(funcname + ': No rescan of {:s}'.format(filename))

        return stat
        
    def finalize_init(self):
        """ Util function that is called by redvypr after initializing all config (i.e. the configuration from a yaml file)
        """
        funcname = self.__class__.__name__ + '.finalize_init()'
        logger.debug(funcname)
        
    def scan_files_clicked(self):
        """ Scans the selected files from the files list for possible datastreams 
        """
        funcname = self.__class__.__name__ + '.scan_files_clicked()'
        logger.debug(funcname)
        rows = []
        for i in self.inlist.selectionModel().selection().indexes():
            row, column = i.row(), i.column()
            rows.append(row)
    
        self.scan_files(rows)
        
    def scan_files(self,rows):
        """ Scans the selected files from the files list for possible datastreams 
        """
        funcname = self.__class__.__name__ + '.scan_files()'
        logger.debug(funcname)

        for i in rows:
            filename = self.inlist.item(i,0).text()
            stat = self.inspect_data(filename,rescan=False)
            packetitem = QtWidgets.QTableWidgetItem(str(stat['packets_published']))
            self.inlist.setItem(i,1,packetitem)
            tdmin = datetime.datetime.fromtimestamp(stat['t_min'])
            tminstr = str(tdmin)
            tdmax = datetime.datetime.fromtimestamp(stat['t_max'])
            tmaxstr = str(tdmax)
            t_min_item = QtWidgets.QTableWidgetItem(tminstr)
            t_max_item = QtWidgets.QTableWidgetItem(tmaxstr)
            self.inlist.setItem(i,2,t_min_item)
            self.inlist.setItem(i,3,t_max_item)
            
        self.inlist.resizeColumnsToContents()
            
    def rem_files(self):
        """ Remove the selected files from the files list
        """
        funcname = self.__class__.__name__ + '.rem_files()'
        logger.debug(funcname)
        rows = []
        for i in self.inlist.selectionModel().selection().indexes():
            row, column = i.row(), i.column()
            rows.append(row)
        
        rows = list(set(rows))
        rows.sort(reverse=True)  
        for i in rows:
            self.device.custom_config['files'].pop(i)
            
        self.update_filenamelist() 

    def add_files(self):
        """ Opens a dialog to choose file to add
        """
        funcname = self.__class__.__name__ + '.add_files()'
        logger.debug(funcname)
        filenames, _ = QtWidgets.QFileDialog.getOpenFileNames(self,"Rawdatafiles","","redvypr raw (*.redvypr_yaml);;All Files (*)")
        for f in filenames: 
            self.device.custom_config['files'].append(f)
            
        
        self.update_filenamelist()
        self.inlist.sortItems(0, QtCore.Qt.AscendingOrder)
            
    def update_filenamelist(self):
        """ Update the filetable 
        """
        funcname = self.__class__.__name__ + '.add_files()'
        logger.debug(funcname)
        self.inlist.clear()
        self.inlist.setHorizontalHeaderLabels(self.__filelistheader__)
        nfiles = len(self.device.custom_config['files'])
        self.inlist.setRowCount(nfiles)

        rows = []        
        for i,f in enumerate(self.device.custom_config['files']):
            item = QtWidgets.QTableWidgetItem(f)
            self.inlist.setItem(i,0,item)
            rows.append(i)
            
            
        self.scan_files(rows)
        self.inlist.resizeColumnsToContents()
        
    def resort_files(self):
        """ Resorts the files in config['files'] according to the sorting in the table
        """
        files_new = []
        nfiles = len(self.device.custom_config['files'])
        for i in range(nfiles):
            filename = self.inlist.item(i,0).text()
            files_new.append(filename)
            
        self.device.custom_config['files'] = files_new
        
    def con_clicked(self):
        funcname = self.__class__.__name__ + '.con_clicked():'
        logger.debug(funcname)
        button = self.sender()
        if(button == self.adddeviceinbtn):
            self.connect.emit(self.device)
            
    def start_clicked(self):
        funcname = self.__class__.__name__ + '.start_clicked():'
        logger.debug(funcname)
        button = self.sender()
        if button.isChecked():
            logger.debug(funcname + "button pressed")
            self.resort_files()
            loop = self.loop_checkbox.isChecked()
            # Loop
            self.device.custom_config['loop']    = loop
            # Speedup
            self.device.custom_config['speedup'] = float(self.speedup_edit.text())
            self.device.thread_start()
        else:
            logger.debug(funcname + 'button released')
            self.device.thread_stop()

            
    def update_buttons(self):
            """ Updating all buttons depending on the thread status (if its alive, graying out things)
            """

            status = self.device.get_thread_status()
            thread_status = status['thread_running']

            # Running
            if(thread_status):
                self.startbtn.setText('Stop')
                self.startbtn.setChecked(True)
                for w in self.config_widgets:
                    w.setEnabled(False)
            # Not running
            else:
                self.startbtn.setText('Start')
                for w in self.config_widgets:
                    w.setEnabled(True)
                    
                # Check if an error occured and the startbutton 
                if(self.startbtn.isChecked()):
                    self.startbtn.setChecked(False)
                #self.conbtn.setEnabled(True)


class displayDeviceWidget(QtWidgets.QWidget):
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout          = QtWidgets.QVBoxLayout(self)
        hlayout         = QtWidgets.QHBoxLayout()
        self.device     = device
        self.text       = QtWidgets.QPlainTextEdit(self)
        self.text.setReadOnly(True)
        self.filelab= QtWidgets.QLabel("File: ")
        self.byteslab   = QtWidgets.QLabel("Bytes written: ")
        self.packetslab = QtWidgets.QLabel("Packets written: ")
        self.text.setMaximumBlockCount(10000)
        hlayout.addWidget(self.byteslab)
        hlayout.addWidget(self.packetslab)
        layout.addWidget(self.filelab)        
        layout.addLayout(hlayout)
        layout.addWidget(self.text)
        #self.text.insertPlainText("hallo!")
        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.update)
        self.statustimer.start(500)

    def update_data(self):
        statusqueue = self.device.statusqueue
        while (statusqueue.empty() == False):
            try:
                data = statusqueue.get(block=False)
            except:
                break

            self.text.insertPlainText(str(data) + '\n')

        #self.byteslab.setText("Bytes written: {:d}".format(data['bytes_written']))
        #self.packetslab.setText("Packets written: {:d}".format(data['packets_written']))
        #self.text.insertPlainText(str(data['data']))
        



"""
sensor
function binary2raw -> dict with datakey (parameter)
generic_sensor (with optional calibrations)

"""
import logging
import sys
import typing
import pydantic
import json
import yaml
import re
import struct
import time
import datetime
import numbers
from PyQt5 import QtWidgets, QtCore, QtGui
import qtawesome
from redvypr.data_packets import check_for_command
from  redvypr.data_packets import create_datadict as redvypr_create_datadict
#import redvypr.config as redvypr_config
from redvypr.redvypr_address import RedvyprAddress, RedvyprAddressStr
from redvypr.device import RedvyprDevice
import redvypr.files as redvypr_files
import redvypr.widgets.standard_device_widgets
from redvypr.devices.plot.XYplotWidget import XYPlotWidget, configXYplot
from redvypr.widgets.pydanticConfigWidget import pydanticConfigWidget, datastreamMetadataWidget
from redvypr.devices.sensors.calibration.calibration_models import calibration_models, calibration_NTC
from redvypr.devices.sensors.csvsensors.sensorWidgets import sensorCoeffWidget, sensorConfigWidget
from redvypr.gui import iconnames
from .sensor_definitions import Sensor, BinarySensor, predefined_sensors
from .calibrationWidget import sensorCalibrationsWidget

_icon_file = redvypr_files.icon_file

redvypr_devicemodule = True

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('generic_sensor')
logger.setLevel(logging.DEBUG)

class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = True
    subscribes: bool = True
    description: str = 'Processing and conversion of raw data to unit using calibration models'

class DeviceCustomConfig(pydantic.BaseModel):
    sensors: typing.List[typing.Annotated[typing.Union[Sensor, BinarySensor], pydantic.Field(discriminator='sensortype')]]\
        = pydantic.Field(default=[], description = 'List of sensors')
    #calibrations: list = pydantic.Field(default=[])
    calibrations: typing.List[typing.Annotated[typing.Union[*calibration_models], pydantic.Field(discriminator='calibration_type')]] = pydantic.Field(default=[])
    calibration_files: list = pydantic.Field(default=[])



def start(device_info, config = None, dataqueue = None, datainqueue = None, statusqueue = None):
    funcname = __name__ + '.start():'
    logger.debug(funcname)
    #print('Got config serialized', config)
    config = DeviceCustomConfig.model_validate(config)
    #print('Got config', config)
    for sensor in config.sensors:
        logger.debug('Creating metadata packet for sensor {}'.format(sensor.name))
        metadata_datapacket = sensor.create_metadata_datapacket()
        #print('Metadata datapacket',metadata_datapacket)
        sensor.add_all_calibrations(config.calibrations)
        if metadata_datapacket is not None:
            dataqueue.put(metadata_datapacket)

    #splitter = BinaryDataSplitter(config.sensors)
    #print('Splitter',splitter)
    while True:
        data = datainqueue.get(block = True)
        if(data is not None):
            command = check_for_command(data, thread_uuid=device_info['thread_uuid'])
            if (command == 'stop'):
                logger.debug('Got a command: {:s}'.format(str(data)))
                logger.debug('Command is for me: {:s}'.format(str(command)))
                break

            for sensor in config.sensors:
                #print('Sensor',sensor)
                #print('Type sensor', type(sensor))
                #print('data',data)
                # Checking for calibrations
                try:
                    sensordata = sensor.datapacket_process(data)
                except:
                    logger.info('Could not decode datapacket',exc_info=True)
                #print('Sensordata',sensordata)
                if type(sensordata) is list: # List means that there was a valid packet
                    for data_packet in sensordata:
                        #print('Publishing data_packet',data_packet)
                        dataqueue.put(data_packet)


class Device(RedvyprDevice):
    """
    generic_sensor device
    """
    newcalibration_signal = QtCore.pyqtSignal()  # Signal for a new calibration
    def __init__(self, **kwargs):
        """
        """
        funcname = __name__ + '__init__()'
        super(Device, self).__init__(**kwargs)
        self.calfiles_processed = []


    def update_subscriptions(self):
        funcname = __name__ + '.update_subscriptions():'
        logger.debug(funcname)
        self.unsubscribe_all()
        for sensor in self.custom_config.sensors:
            self.subscribe_address(sensor.datastream)

    def add_sensor(self, sensor):
        flag_name_new = True
        for sen in self.custom_config.sensors:
            if sen.name == sensor.name:
                flag_name_new = False
                break

        if flag_name_new:
            logger.debug('Adding sensor {}'.format(sensor.name))
            self.custom_config.sensors.append(sensor)
            self.update_subscriptions()
            self.config_changed_signal.emit()
        else:
            raise('Sensor name exists already')

    def remove_sensor(self, sensor):
        isensor = self.custom_config.sensors.index(sensor)
        if isensor >=0:
            logger.debug('Removing sensor {}'.format(sensor.name))
            self.custom_config.sensors.pop(isensor)
            self.update_subscriptions()
            self.config_changed_signal.emit()

    def add_calibration(self, calibration):
        """
        Adds a calibration to the calibration list, checks before, if the calibration exists
        calibration: calibration model
        """
        flag_new_calibration = True
        calibration_json = json.dumps(calibration.model_dump_json())
        for cal_old in self.custom_config.calibrations:
            if calibration_json == json.dumps(cal_old.model_dump_json()):
                flag_new_calibration = False
                break

        if flag_new_calibration:
            print('Sending new calibration signal')
            self.custom_config.calibrations.append(calibration)
            self.newcalibration_signal.emit()
            return True
        else:
            logger.warning('Calibration exists already')
            return False

    def read_calibration_file(self, fname):
        """
        Open and reads a calibration file, it will as well determine the type of calibration and call the proper function

        """
        funcname = __name__ + '.read_calibration_file():'
        logger.debug(funcname + 'Opening file {:s}'.format(fname))
        try:
            f = open(fname)
            safe_load = False
            if safe_load:
                loader = yaml.SafeLoader
            else:
                loader = yaml.CLoader

            data = yaml.load(f,Loader=loader)
            # print('data',data)
            if 'structure_version' in data.keys(): # Calibration model
                logger.debug(funcname + ' Version {} pydantic calibration model dump'.format(data['structure_version']))
                for calmodel in calibration_models: # Loop over all calibration models definded in sensor_calibrations.py
                    try:
                        calibration = calmodel.model_validate(data)
                        print('Calibration', calibration)
                        return calibration
                    except:
                        pass

            elif 'sn' in data.keys():
                calibration = {}
                calibration['sn'] = {}
                logger.debug(funcname + ' Version 2 file')
                for sn in data['sn'].keys():  # Add the filename the data was loaded from
                    try:
                        calibration['sn'][sn]
                    except:
                        calibration['sn'][sn] = {}
                        #calibration['sn'][sn]['coeff_files'] = []

                    #calibration['sn'][sn]['coeff_files'].append(fname)
                    calibration['sn'][sn].update(data['sn'][sn])
                    calibration['sn'][sn]['original_file'] = fname
                #    data['sn'][sn].filenames = []
                #    data['sn'][sn].filenames.append(fname)
                # self.calibration['sn'].update(data['sn'])
                return calibration

            elif 'calibration_HF_SI' in data.keys(): # Version 1 file
                logger.debug(funcname + ' Version 1 file')

                sn = data['manufacturer_sn']
                if len(sn) ==0:
                    logger.debug(funcname + ' No serial number')
                else:
                    calibration = calibration_HF(sn=sn,coeff=data['calibration_HF_SI'],date=data['calibration_date'],sensor_model = data['series'])
                    #sn:
                    #    "9299":
                    #    model: F - 005 - 4
                    #    coeffs:
                    #        HF:
                    #        coeff: 8.98
                    #        parameter: HF
                    #        unit: W m-2
                    #        unitraw: mV
                    #        date: 1970 - 01 - 01_00 - 00 - 00
                    return calibration
                    if False:
                        calibration['sn'][sn] = {}
                        calibration['sn'][sn]['model'] = data['series']
                        calibration['sn'][sn]['original_file'] = fname
                        calibration['sn'][sn]['sn'] = sn
                        calibration['sn'][sn]['parameter'] = {}
                        calibration['sn'][sn]['parameter']['HF'] = {}
                        calibration['sn'][sn]['parameter']['HF']['coeff'] = data['calibration_HF_SI']
                        calibration['sn'][sn]['parameter']['HF']['unit'] = 'W m-2'
                        calibration['sn'][sn]['parameter']['HF']['unitraw'] = 'mW'
                        calibration['sn'][sn]['parameter']['HF']['date'] = data['calibration_date']

        except Exception as e:
            logger.exception(e)
            return None

    def add_calibration_file(self, calfile, reload=True):
        funcname = __name__ + 'add_calibration_file():'
        logger.debug(funcname)
        calfiles_tmp = list(self.custom_config.calibration_files)
        if (calfile in calfiles_tmp) and reload:
            #print('Removing file first')
            self.rem_calibration_file(calfile)

        calfiles_tmp = list(self.custom_config.calibration_files)
        #print('Hallo',calfiles_tmp)
        if calfile not in calfiles_tmp:
            #print('Adding file 2',calfile)
            self.custom_config.calibration_files.append(calfile)
        else:
            logger.debug(funcname + ' File is already listed')

        self.process_calibrationfiles()

    def rem_calibration_file(self, calfile):
        funcname = __name__ + 'rem_calibration_file():'
        logger.debug(funcname)
        calfiles_tmp = list(self.custom_config.calibration_files)
        if calfile in calfiles_tmp:
            calibration = self.read_calibration_file(calfile)
            calibration_json = json.dumps(calibration.model_dump_json())
            for cal_old in self.custom_config.calibrations:
                # Test if the calibration is existing
                if calibration_json == json.dumps(cal_old.model_dump_json()):
                    logger.debug(funcname + ' Removing calibration')
                    self.custom_config.calibration_files.remove(calfile)
                    self.custom_config.calibrations.remove(cal_old)
                    self.calfiles_processed.remove(calfile)
                    return True

    def process_calibrationfiles(self):
        funcname = __name__ + '.process_calibrationfiles()'
        logger.debug(funcname)
        fnames = []

        for fname in self.custom_config.calibration_files:
            fnames.append(str(fname))

        self.coeff_filenames = fnames

        for fname in fnames:
            if fname not in self.calfiles_processed:
                logger.debug(funcname + ' reading file {:s}'.format(fname))
                calibration = self.read_calibration_file(fname)
                if calibration is not None:
                    flag_add = self.add_calibration(calibration)
                    if flag_add:
                        self.calfiles_processed.append(fname)
            else:
                logger.debug(funcname + ' file {:s} already processed'.format(fname))


        #print(self.calibration['sn'].keys())
        #self.logger_autocalibration()


class RedvyprDeviceWidget(redvypr.widgets.standard_device_widgets.RedvyprDeviceWidget_startonly):
    def __init__(self,*args,**kwargs):
        super().__init__(*args, **kwargs)
        self.devicedisplaywidget = displayDeviceWidget(device=self.device)
        icon = qtawesome.icon(iconnames['settings'])
        self.sensorshow_combo = QtWidgets.QComboBox()
        self.settings_button = QtWidgets.QPushButton('Settings')
        self.settings_button.setIcon(icon)
        self.settings_button.clicked.connect(self.settings_clicked)
        self.layout.removeWidget(self.buttons_widget)
        self.layout.removeWidget(self.killbutton)
        self.killbutton.hide()
        self.layout.addWidget(self.devicedisplaywidget,0,0)
        self.layout.addWidget(self.buttons_widget,1,0)
        self.layout_buttons.addWidget(self.sensorshow_combo, 0, 4)
        self.layout_buttons.addWidget(self.settings_button,0,5)

        self.device.config_changed_signal.connect(self.update_sensorcombo)
        self.update_sensorcombo()
        self.sensorshow_combo.currentIndexChanged.connect(self.sensorshow_changed)

    def settings_clicked(self):
        self.settings_widget = QtWidgets.QWidget()
        self.sensortable_status = sensorTableWidget(device=self.device)
        self.settings_widget_layout = QtWidgets.QVBoxLayout(self.settings_widget)
        self.initdevicewidget = initDeviceWidget(device=self.device, redvypr=self.redvypr)
        self.settings_widget_layout.addWidget(self.sensortable_status)
        self.settings_widget_layout.addWidget(self.initdevicewidget)
        self.settings_widget.show()

    def sensorshow_changed(self):
        funcname = __name__ + '.sensorshow_changed():'
        logger.debug(funcname)
        sensorname = self.sensorshow_combo.currentText()
        print('Sensorname', sensorname)
        for irow, sensor in enumerate(self.device.custom_config.sensors):
            if sensorname == sensor.name:
                sensorwidget = sensor.__sensorwidget__
                self.devicedisplaywidget.sensor_show(sensorwidget)

    def update_sensorcombo(self):
        funcname = __name__ + '.update_sensorcombo():'
        logger.debug(funcname)
        print('Updating...')
        self.sensorshow_combo.clear()
        for irow,sensor in enumerate(self.device.custom_config.sensors):
            name = sensor.name
            self.sensorshow_combo.addItem(name)


    def update_data(self, data):
        self.devicedisplaywidget.update_data(data)



class initDeviceWidget(redvypr.widgets.standard_device_widgets.redvypr_deviceInitWidget):
    def __init__(self, *args, **kwargs):
        funcname = __name__ + '__init__():'
        logger.debug(funcname)
        super().__init__(*args, **kwargs)

        self.sensorwidget = QtWidgets.QWidget()
        self.sensorwidget_layout = QtWidgets.QVBoxLayout(self.sensorwidget)

        self.sensortable = QtWidgets.QTableWidget()
        self.sensortable.setRowCount(1)

        self.sensor_addbutton = QtWidgets.QPushButton('Add sensor')
        self.sensor_addbutton.clicked.connect(self.add_sensor_clicked)
        self.sensorwidget_layout.addWidget(self.sensortable)
        self.sensorwidget_layout.addWidget(self.sensor_addbutton)

        self.sensor_calibrations = QtWidgets.QPushButton('Calibrations')
        self.sensor_calibrations.clicked.connect(self.calibrations_clicked)
        self.sensorwidget_layout.addWidget(self.sensor_calibrations)

        self.layout.removeWidget(self.label)
        self.label.setParent(None)

        self.layout.addWidget(self.sensorwidget, 0,0,1,-1)
        self.update_sensortable()
        # Add the widgets to the config widget, that enables/disables them when thread is running
        self.config_widgets.append(self.sensortable)
        self.config_widgets.append(self.sensor_addbutton)
        self.config_widgets.append(self.sensor_calibrations)

    def calibrations_clicked(self):
        self.calibrationsWidget = sensorCalibrationsWidget(calibrations=self.device.custom_config.calibrations, redvypr_device = self.device)
        self.calibrationsWidget.show()

    def update_sensortable(self):
        funcname = __name__ + '.update_sensortable():'
        nsensors = len(self.device.custom_config.sensors)
        self.sensortable.clear()
        self.sensortable.setRowCount(nsensors)
        colheaders = ['Name', 'Datastream', 'Configure', 'Remove']
        self.sensortable.setColumnCount(4)
        self.sensortable.setHorizontalHeaderLabels(colheaders)
        icol_name = 0
        icol_subscribe = 1
        icol_config = 2
        icol_delete = 3
        for irow,sensor in enumerate(self.device.custom_config.sensors):
            name = sensor.name
            item_name = QtWidgets.QTableWidgetItem(name)
            self.sensortable.setItem(irow,icol_name,item_name)
            datastream = str(sensor.datastream)
            item_datastream = QtWidgets.QTableWidgetItem(datastream)
            self.sensortable.setItem(irow, icol_subscribe, item_datastream)

            config_icon = qtawesome.icon('fa.navicon')
            item_config = QtWidgets.QPushButton(config_icon,'Config')
            item_config.__sensor__ = sensor
            item_config.clicked.connect(self.sensor_config_clicked)
            self.sensortable.setCellWidget(irow, icol_config, item_config)

            bin_icon = qtawesome.icon('ri.delete-bin-5-fill')
            item_del = QtWidgets.QPushButton(bin_icon,'Remove')
            item_del.__sensor__ = sensor
            item_del.clicked.connect(self.sensor_remove_clicked)
            self.sensortable.setCellWidget(irow, icol_delete, item_del)

        self.sensortable.resizeColumnsToContents()

    def sensor_config_clicked(self):
        sensor = self.sender().__sensor__
        self.sensor_config_widget = pydanticConfigWidget(sensor,redvypr=self.device.redvypr)
        self.sensor_config_widget.config_editing_done.connect(self.device.update_subscriptions)
        self.sensor_config_widget.show()

    def sensor_remove_clicked(self):
        sensor = self.sender().__sensor__
        self.device.remove_sensor(sensor)
        self.update_sensortable()

    def add_sensor_clicked(self):
        self.add_sensor_widget = QtWidgets.QWidget()
        self.add_sensor_widget.setWindowTitle('Add sensor')
        self.add_sensor_widget_layout = QtWidgets.QGridLayout(self.add_sensor_widget)
        # Standard sensors
        self.sensor_standard_combo = QtWidgets.QComboBox()
        self.sensor_standard_combo.addItem('Sensor')
        self.sensor_standard_combo.addItem('BinarySensor')
        self.sensor_standard_combo.setCurrentIndex(1)
        #self.sensor_standard_combo.currentIndexChanged.connect(self.add_sensor_combo_changed)
        self.sensor_standard_choose = QtWidgets.QPushButton('Choose sensor')
        self.sensor_standard_choose.clicked.connect(self.add_sensor_combo_changed)
        self.add_sensor_widget_layout.addWidget(self.sensor_standard_combo, 0, 0)
        self.add_sensor_widget_layout.addWidget(self.sensor_standard_choose, 1, 0)
        # Predefined sensors
        self.sensor_combo = QtWidgets.QComboBox()
        for sensor in predefined_sensors:
            self.sensor_combo.addItem(sensor.name)
        #self.sensor_combo.currentIndexChanged.connect(self.add_sensor_combo_changed)
        self.sensor_predefined_choose = QtWidgets.QPushButton('Choose sensor')
        self.sensor_predefined_choose.clicked.connect(self.add_sensor_combo_changed)
        self.add_sensor_widget_layout.addWidget(self.sensor_combo,0,1)
        self.add_sensor_widget_layout.addWidget(self.sensor_predefined_choose, 1, 1)

        #self.sensor_standard_combo.setCurrentIndex(0)
        self.sensor_standard_choose.clicked.emit()
        self.add_sensor_widget.show()

    def add_sensor_combo_changed(self):
        #combo = self.sender()
        button = self.sender()
        if button == self.sensor_predefined_choose:
            combo = self.sensor_combo
        else:
            combo = self.sensor_standard_combo
        sensorname = combo.currentText()
        sensorindex = combo.currentIndex()
        # Redefined sensor
        if combo == self.sensor_combo:
            sensor = predefined_sensors[sensorindex].model_copy()
        else:
            if sensorindex == 0:
                sensor = Sensor()
            else:
                sensor = BinarySensor()

        try:
            self.add_sensor_widget_layout.removeWidget(self.add_sensor_config_widget)
            self.add_sensor_config_widget.setParent(None)
        except:
            pass

        self.__sensor_to_add__ = sensor
        self.add_sensor_config_widget = pydanticConfigWidget(self.__sensor_to_add__, redvypr=self.device.redvypr, close_after_editing=False)
        self.add_sensor_config_widget.config_editing_done.connect(self.add_sensor_applied)
        self.add_sensor_widget_layout.addWidget(self.add_sensor_config_widget,2,0,1,2)
        #self.config_widget.showMaximized()
        #print('Sensorname',sensorname,'sensorindex',sensorindex)
        #print('Sensor',sensor)

    def add_sensor_applied(self):
        funcname = __name__ + '.add_sensor_applied():'
        logger.debug(funcname)
        try:
            self.device.add_sensor(self.__sensor_to_add__)
            self.add_sensor_widget.close()
            self.update_sensortable()
        except:
            logger.info('Could not add sensor', exc_info=True)
        #self.device.custom_config.sensors.append(self.__sensor_to_add__)







class displayDeviceWidget(QtWidgets.QWidget):
    """ Widget is displaying incoming data as text
    """

    def __init__(self, device=None, tabwidget=None):
        """
        device [optional]
        tabwidget [optional]

        """
        funcname = __name__ + '.__init__()'
        super(QtWidgets.QWidget, self).__init__()
        layout = QtWidgets.QVBoxLayout(self)
        self.layout = layout
        self.tabwidget = tabwidget
        self.device = device
        self.sensorwidget = None # The widget showing the sensor data
        self.device.config_changed_signal.connect(self.update_sensorwidgets)
        self.update_sensorwidgets()
        try:
            sensorwidget = self.device.custom_config.sensors[0].__sensorwidget__
            self.sensor_show(sensorwidget)
        except:
            logger.info('Could not show widget',exc_info=True)

    def update_sensorwidgets(self):
        funcname = __name__ + '.update_sensorwidgets():'
        nsensors = len(self.device.custom_config.sensors)
        for irow, sensor in enumerate(self.device.custom_config.sensors):
            # Create a sensorwidget, if not existing
            try:
                sensor.__sensorwidget__
            except:
                sensor.__sensorwidget__ = SensorWidget(sensor=sensor, redvypr_device=self.device)

    def sensor_show_clicked(self):
        funcname = __name__ + '.sensor_show_clicked():'
        logger.debug(funcname)
        sensor = self.sender().__sensor__
        self.sensor_show(sensor.__sensorwidget__)

    def sensor_show(self,sensorwidget):
        try:
            self.layout.removeWidget(self.sensorwidget)
            self.sensorwidget.hide()
            # sensorwidget_old = self.splitter.replaceWidget(1,self.sensorwidget)
            # sensorwidget_old.hide()
        except:
            logger.debug('Could not remove widget', exc_info=True)
            # self.splitter.addWidget(self.sensorwidget)

        self.sensorwidget = sensorwidget
        self.layout.addWidget(self.sensorwidget)
        self.sensorwidget.show()

    def sensor_plot_clicked(self):
        print('Plot')
    def thread_status(self, status):
        """ This function is regularly called by redvypr whenever the thread is started/stopped
        """
        pass
        # self.update_buttons(status['threadalive'])



    def update_data(self, data):
        """
        """
        funcname = __name__ + '.update_data()'
        tnow = time.time()
        for sensor in self.device.custom_config.sensors:
            try:
                sensor.__sensorwidget__.update_data(data)
            except:
                logger.info('Could not update',exc_info=True)



class SensorWidget(QtWidgets.QWidget):
    def __init__(self, *args, sensor=None, redvypr_device=None, **kwargs):
        funcname = __name__ + '__init__():'
        #logger.debug(funcname)
        self.packets_received = 0
        self.packets_for_me = 0
        self.packets_for_me_old = 0
        self.sensor = sensor
        self.device = redvypr_device
        #print('Sensor',self.sensor)
        super().__init__(*args, **kwargs)
        self.layout = QtWidgets.QHBoxLayout(self)
        #self.label = QtWidgets.QLabel(self.sensor.name)
        self.sensor_address = RedvyprAddress('/d:{}'.format(sensor.name))
        # Table showing the different packetids, which are used to distinguis different sensors
        self.packetidtable = QtWidgets.QTableWidget()
        self.packetidtable.setRowCount(0)
        self.packetidtable.setColumnCount(2)
        self.packetidtable.cellClicked.connect(self.packetid_changed)
        self.datatable = QtWidgets.QTableWidget()
        self.datatable.setRowCount(2)
        self.datatable.setColumnCount(4)
        colheaders = ['Datakey', 'Data', 'Unit', 'XY-Plot']
        self.datatable.setHorizontalHeaderLabels(colheaders)
        self.datakey_items = {} # Dictionary with the tablewidgetitems
        self.datakey_units = {} # Dictionary with the units
        self.datakey_plot = {}# Dictionary with the plot widgets
        self.datakey_plot_data = {}  # Dictionary with the data for the plots
        self.bufferlen = 1000
        self.last_packet = {}  # Dictionary with the plot widgets
        self.packetids = []  # Dictionary with the plot widgets
        self.icol_key = 0
        self.icol_data = 1
        self.icol_unit = 2
        self.icol_plot = 3
        item_key_tstr = QtWidgets.QTableWidgetItem('Time')
        item_key_tustr = QtWidgets.QTableWidgetItem('Time (unix)')
        #self.datatable.setItem(0, self.icol_data, item_tustr)
        #self.datatable.setItem(1, self.icol_data, item_tstr)
        self.datatable.setItem(0, self.icol_key, item_key_tustr)
        self.datatable.setItem(1, self.icol_key, item_key_tstr)

        #self.layout.addWidget(self.label,1)
        self.layout.addWidget(self.packetidtable,1)

        self.layout.addWidget(self.datatable,4)
        self.__packetid_selected = None

    def packetid_changed(self,row,column):
        funcname = __name__ + '.packetid_changed():'
        logger.debug(funcname)
        #print('Packetid changed')
        try:
            packetid = self.packetidtable.item(row,column).text()
            self.__packetid_selected = packetid
            data = self.last_packet[packetid]
            #print('Updating with',packetid)
            #print('Data',data)
            self.datakey_items = {}  # Dictionary with the tablewidgetitems
            self.datakey_units = {}  # Dictionary with the units
            self.__packetid_selected = packetid
            #self.datatable.clear()
            self.datatable.setRowCount(2)
            self.update_data(data)
        except:
            logger.info('Could not change packeid:',exc_info=True)
            packetid = None


        #print(funcname, row, column, packetid)

    def sensor_plot_clicked(self):
        funcname = __name__ + '.sensor_plot_clicked():'
        logger.debug(funcname)
        # No XY-Plot yet, create one
        if self.sender().__xyplot__ is None:
            packetid = self.sender().__packetid__
            k = self.sender().__k__
            logger.debug(funcname + 'Plot clicked, creating XY-Plotwidget')
            logger.debug(funcname + 'sensor address: {} packetid: {}, k: {}'.format(self.sensor_address,packetid,k))
            config_plot = configXYplot(automatic_subscription=False)
            self.datakey_plot[packetid][k] = XYPlotWidget(config=config_plot, add_line=False,
                                                          redvypr_device=self.device)
            yaddr = RedvyprAddress(self.sensor_address, datakey=k, packetid=packetid)

            logger.debug(funcname + 'yaddr: {}'.format(yaddr))
            self.datakey_plot[packetid][k].add_line(yaddr)
            self.sender().__xyplot__ = self.datakey_plot[packetid][k]
            self.datakey_plot[packetid][k].__item_plot__ = self.sender()
            self.datakey_plot[packetid][k].closing.connect(self.__xyplot_closed__)
            windowtitle = 'Plot {} {}'.format(packetid, k)
            self.sender().__xyplot__.setWindowTitle(windowtitle)
            # Adding the buffer data
            buffer_tmp = self.datakey_plot_data[packetid]
            logger.debug(funcname + 'Updating plot with buffer data')
            for data_tmp in buffer_tmp:
                self.datakey_plot[packetid][k].update_plot(data_tmp)


        self.sender().__xyplot__.show()

    def __xyplot_closed__(self):
        funcname = __name__ + '.__xyplot_closed__():'
        logger.debug(funcname)
        xyplot = self.sender()
        item_plot = xyplot.__item_plot__
        packetid = item_plot.__packetid__
        k = item_plot.__k__
        self.datakey_plot[packetid][k] = None
        item_plot.__xyplot__ = None
        xyplot.deleteLater()
        xyplot.setParent(None)
        logger.debug(funcname + 'Closed')

    def __metadata_clicked(self):
        button = self.sender()
        filter_address = RedvyprAddress(packetid='==' + button.__metadata_raddress__.packetid)

        packetid = button.__packetid__
        self.__metawidget = datastreamMetadataWidget(redvypr=self.device.redvypr, device=self.device,
                                                     filter_include=[filter_address])
        self.__metawidget.show()

    def update_data(self, data):
        funcname = __name__ + '.update_data():'
        logger.debug(funcname + 'Got data for sensor {}'.format(self.sensor.name))
        #print('Fresh data:',data,type(data))
        if data in self.sensor_address:
            #print(funcname + ' Datapacket fits, Processing data')
            rdata = redvypr.data_packets.Datapacket(data)
            # Packets are sorted according to the packetid, as this is including the serialnumber
            packetid = rdata.address.packetid
            #print('Got data',rdata)
            #print('Packetid',packetid)

            # Try to update plotdata
            try:
                self.datakey_plot_data[packetid]
            except:
                self.datakey_plot_data[packetid] = []

            self.datakey_plot_data[packetid].append(data)
            if len(self.datakey_plot_data[packetid]) > self.bufferlen:
                self.datakey_plot_data[packetid].pop(0)
            # Check if the packetid has been seen before
            if packetid not in self.packetids and len(packetid) > 0 and (rdata.address.packetidexpand == False):
                self.packetids.append(packetid)
                self.packetidtable.clear()
                self.packetidtable.setRowCount(len(self.packetids))
                self.packetidtable.setHorizontalHeaderLabels(['Packetids for sensor {}'.format(self.sensor.name),'Metadata'])
                for i,p in enumerate(self.packetids):
                    packetiditem = QtWidgets.QTableWidgetItem(p)
                    packetiditem.setFlags(packetiditem.flags() & ~QtCore.Qt.ItemIsEditable)
                    self.packetidtable.setItem(i,0,packetiditem)
                    item_metadata = QtWidgets.QPushButton('Metadata')
                    item_metadata.__packetid__ = packetid
                    item_metadata.__metadata_raddress__ = RedvyprAddress(rdata.address)
                    item_metadata.clicked.connect(self.__metadata_clicked)
                    # self.datakey_plot[packetid][k].__item_plot__ = item_plot
                    icol_metadata = 1
                    self.packetidtable.setCellWidget(i, icol_metadata, item_metadata)
                    if p == self.__packetid_selected:
                        self.packetidtable.setCurrentCell(i, 0)

                self.packetidtable.resizeColumnsToContents()
                if self.__packetid_selected is None:
                    self.__packetid_selected = p
                    self.packetidtable.setCurrentCell(i, 0)

            #print('Packetid',packetid,self.packetids)
            keys = rdata.datakeys(expand=True, return_type='list')
            # Update the last packet dictionary
            if 't' in keys:
                self.last_packet[packetid] = rdata
            # Updating the metadata
            if True: # TODO, this should be done based on a signal notifying that the metadata has changed
                logger.debug('Updating keyinfo')
                # This should be done with device.get_metadata
                #for k in data['_keyinfo'].keys():
                for k in keys:
                    try:
                        self.datakey_units[k]
                    except:
                        raddr_tmp = RedvyprAddress(data,datakey=k)
                        logger.debug(funcname + 'Trying to get metadata for address {}'.format(raddr_tmp))
                        metadata = self.device.get_metadata(raddr_tmp)
                        #print(funcname + ' Got Metadata ...', metadata)
                        try:
                            self.datakey_units[k] = metadata['unit']
                        except:
                            self.datakey_units[k] = None

                    # Update the plot first or create new plot widget
                    if 't' in keys and k is not 't':
                        #print('Packetid',packetid,k)
                        try:
                            self.datakey_plot_data[packetid][k]
                        except:
                            self.datakey_plot_data[packetid][k] = []

                        self.datakey_plot_data[packetid][k].append()
                        # Try to update plots
                        try:
                            self.datakey_plot[packetid]
                        except:
                            self.datakey_plot[packetid] = {}
                        try:
                            xyplot = self.datakey_plot[packetid][k]
                            if xyplot is not None:
                                self.datakey_plot[packetid][k].update_plot(data)
                        except:
                            logger.info('Could not update plot', exc_info=True)
                            logger.debug(funcname + 'Adding XY-Plot (None)')
                            self.datakey_plot[packetid][k] = None

            # Updating the raw data
            flag_packetid = (packetid == self.__packetid_selected)
            #print('Hallo', 'Packetid', packetid, 'Selected', self.__packetid_selected, flag_packetid)
            if 't' in keys:
                self.packets_for_me += 1
            if 't' in keys and flag_packetid:
                #print('Correct packetid, updating table')
                t = keys.remove('t')
                # Display always the time
                tdata = datetime.datetime.utcfromtimestamp(data['t'])
                tudatastr = str(data['t'])
                tdatastr = tdata.strftime('%Y-%m-%d %H:%M:%S.%f')
                try:
                    self.datakey_items['t'][0]
                except:
                    item_tstr = QtWidgets.QTableWidgetItem('NA')
                    item_tustr = QtWidgets.QTableWidgetItem('NA')
                    item_tustr.setText(tudatastr)
                    item_tstr.setText(tdatastr)
                    self.datakey_items = {}
                    self.datakey_items['t'] = [item_tustr, item_tstr]
                    self.datatable.setItem(0, self.icol_data, item_tustr)
                    self.datatable.setItem(1, self.icol_data, item_tstr)

                for k in keys:
                    if True:
                        # Datastring to be displayed
                        datastr = str(rdata[k]) # This can be done more fancy
                        if k not in self.datakey_items:
                            nrows = self.datatable.rowCount()
                            logger.debug(funcname + 'Creating new item {}'.format(nrows))
                            datakeyitem = QtWidgets.QTableWidgetItem(k)
                            dataitem = QtWidgets.QTableWidgetItem(datastr)
                            try:
                                unit = self.datakey_units[k]
                            except:
                                unit = 'NA'

                            dataunititem = QtWidgets.QTableWidgetItem(unit)
                            #self.datatable.insertRow(nrows+1)
                            self.datatable.setRowCount(nrows + 1)
                            self.datatable.setItem(nrows, self.icol_key, datakeyitem)
                            self.datatable.setItem(nrows, self.icol_data, dataitem)
                            self.datatable.setItem(nrows, self.icol_unit, dataunititem)
                            # Add a XY-Plot, if datatype fits
                            #if isinstance(data[k], numbers.Number):
                            try:
                                xyplot = self.datakey_plot[packetid][k]
                                plot_icon = qtawesome.icon('ph.chart-line-fill')
                                #item_plot = QtWidgets.QPushButton(plot_icon, 'Plot {} {}'.format(packetid, k))
                                item_plot = QtWidgets.QPushButton(plot_icon, 'Plot')
                                item_plot.__xyplot__ = self.datakey_plot[packetid][k]
                                item_plot.__packetid__ = packetid
                                item_plot.__k__ = k
                                item_plot.clicked.connect(self.sensor_plot_clicked)
                                #self.datakey_plot[packetid][k].__item_plot__ = item_plot
                                self.datatable.setCellWidget(nrows, self.icol_plot, item_plot)
                            except:
                                logger.warning('Could not get item_plot',exc_info=True)
                                item_plot = None


                            self.datakey_items[k] = [nrows, datakeyitem, dataitem, dataunititem, item_plot]
                        else:
                            #print('Just updating',datastr,k)
                            # Data
                            # This could be only done less brute force if packetid has been changed ...
                            irow = self.datakey_items[k][0]
                            nrows = self.datatable.rowCount()
                            datakeyitem = self.datakey_items[k][1]
                            dataitem = self.datakey_items[k][2]
                            dataunititem = self.datakey_items[k][3]
                            item_plot = self.datakey_items[k][4]
                            #print('dataitem',dataitem.text())
                            dataitem.setText(datastr)
                            # Unit
                            try:
                                unit = self.datakey_units[k]
                                dataunititem.setText(unit)
                            except:
                                pass
                                #logger.debug('Could not update unit',exc_info=True)


                self.datatable.resizeColumnsToContents()



class sensorTableWidget(QtWidgets.QWidget):
    """ Widget is displaying all defined sensors
    """

    def __init__(self, device=None):
        funcname = __name__ + '.__init__()'
        super(QtWidgets.QWidget, self).__init__()
        layout = QtWidgets.QVBoxLayout(self)
        self.layout = layout
        self.device = device
        # A timer that is regularly calling the device.status function
        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.update_status)
        self.dt_update = 1
        self.sensortable = QtWidgets.QTableWidget()
        self.layout.addWidget(self.sensortable)
        self.device.config_changed_signal.connect(self.update_sensortable)
        # The table columns
        self.icol_name = 0
        self.icol_packet_proc = 1
        self.icol_packet_proc__s = 2
        self.icol_show = 3
        self.icol_plot = 4
        self.update_sensortable()
        self.statustimer.start(self.dt_update * 1000)

    def update_sensortable(self):
        funcname = __name__ + '.update_sensortable():'
        nsensors = len(self.device.custom_config.sensors)
        self.sensortable.clear()
        self.sensortable.setRowCount(nsensors)
        colheaders = ['Name', 'Packets processed', 'Packets/s']
        self.sensortable.setColumnCount(3)
        self.sensortable.setHorizontalHeaderLabels(colheaders)

        for irow, sensor in enumerate(self.device.custom_config.sensors):
            # Create a sensorwidget, if not existing
            try:
                sensor.__sensorwidget__
            except:
                logger.warning('Could not find widget for sensor')
                continue

            name = sensor.name
            item_name = QtWidgets.QTableWidgetItem(name)
            self.sensortable.setItem(irow, self.icol_name, item_name)

            #show_icon = qtawesome.icon('ei.list')
            #item_show = QtWidgets.QPushButton(show_icon, 'Show')
            #item_show.__sensor__ = sensor
            #item_show.clicked.connect(self.sensor_show_clicked)
            #self.sensortable.setCellWidget(irow, self.icol_show, item_show)

            npackets = sensor.__sensorwidget__.packets_for_me
            item_packet = QtWidgets.QTableWidgetItem(str(npackets))
            self.sensortable.setItem(irow, self.icol_packet_proc, item_packet)

            item_packet__s = QtWidgets.QTableWidgetItem(str(0))
            self.sensortable.setItem(irow, self.icol_packet_proc__s, item_packet__s)

        self.sensortable.resizeColumnsToContents()

    def update_status(self):
        """
        """
        funcname = __name__ + 'update_status():'
        try:
            statusdata = self.device.get_status()
            # print(funcname + str(statusdata))
        except:
            logger.debug(funcname,exc_info=True)

        # update the
        for isensor, sensor in enumerate(self.device.custom_config.sensors):
            npackets_old = sensor.__sensorwidget__.packets_for_me_old
            npackets = sensor.__sensorwidget__.packets_for_me
            npackets__s = (npackets - npackets_old)/self.dt_update
            sensor.__sensorwidget__.packets_for_me_old = npackets

            item = self.sensortable.item(isensor, self.icol_packet_proc)
            item.setText(str(npackets))

            item = self.sensortable.item(isensor, self.icol_packet_proc__s)
            item.setText(str(npackets__s))


